#! /usr/bin/env python
"""Create .deb sets from unattended upgrade"""
from __future__ import print_function
from fussy import nbio, twrite
import re, os, time, logging, json
import tarfile
log = logging.getLogger(__name__)
TARGET_MATCHER = re.compile(r"""DestFile:['](?P<filename>[^']+)[']""")
DEPENDENCY_MATCHER = re.compile(r"""(?P<package>[^, ()]*?)[ ]*[(](?P<compare>[<>=]+)[ ]*(?P<version>[^)]*?)[)]""")
DIGIT_FINDER = re.compile( '\d+' )
def parse_dpkg_info( content ):
    versions = {}
    content = content.strip().splitlines()
    current = None
    for line in content:
        for interesting in [
            'Package:', 'Version:', 'Status:', 
            'Depends:','Pre-Depends:',
        ]:
            line = line.strip()
            if line.startswith(interesting):
                value = line[len(interesting):].strip()
                if interesting == 'Package:':
                    if current:
                        versions[current['package']] = current
                    current = dict( package = value )
                elif interesting == 'Version:':
                    version = line.split(':', 1)[1].strip()
                    current['version'] = [ 
                        int(x,10) 
                        for x in DIGIT_FINDER.findall( version )
                    ]
                elif interesting == 'Status:':
                    if 'installed' in value.split():
                        current['installed'] = True 
                    else:
                        current['installed'] = False
                elif interesting in ('Depends:','Pre-Depends:'):
                    value= [
                        x.groupdict() 
                        for x in DEPENDENCY_MATCHER.finditer( value )
                    ]
                    if interesting is 'Depends:':
                        key = 'depends'
                    else:
                        key = 'predepends'
                    current[key] = value
                else:
                    current[interesting.lower().rstrip(':')] = value
    if current:
        versions[current['package']] = current
    return versions

def get_options():
    import argparse
    parser = argparse.ArgumentParser(description='Process unattended upgrade log to produce upgrade set')
    default_file = 'os-upgrades-%s.deb.tar'%(time.strftime('%Y-%m-%d-%H-%M', time.gmtime()))
    parser.add_argument(
        '--target', metavar='FILENAME.deb.tar',
        default = default_file, 
        help="Target directory to watch (must exist), default: %s"%(default_file, ),
    )
    parser.add_argument(
        '--list',  action='store_true', 
        default=False, 
        help = "Just list the files that would be packed, do not do actual packing", 
    )
    parser.add_argument(
        '--exclude', metavar='PACKAGE', 
        dest = 'excluded', 
        action = 'append', 
        help='Packages *not* to include in the resulting tar-file', 
    )
    parser.add_argument(
        '--recalculate', action='store_true', 
        default=False, 
        help='If provided, will run unattended upgrades with dry-run to calculate (and download) the set', 
    )
    return parser

def get_metadata_options():
    import argparse
    parser = argparse.ArgumentParser(description='Extract metadata from .deb files for unattended set')
    parser.add_argument(
        '-m',
        '--merge',
        metavar='FULLPATH.json',
        default=None,
        help='If provided, metadata will be *merged* into existing json file',
    )
    parser.add_argument(
        'filenames',
        metavar='FULLPATH.deb',
        nargs = "+",
    )
    return parser

def recalculate():
    log.warn("Running unattended upgrades, this will take a long time")
    nbio.Process('sudo unattended-upgrades -d --dry-run')()

def filter_exclusions( package, excluded ):
    import fnmatch
    for exclude in excluded:
        if fnmatch.fnmatch( package, exclude ):
            log.info("Excluding %s due to pattern %s", package, exclude)
            return False 
    return True

def get_components(excluded=None):
    result = []
    unique = set()
    for line in nbio.Process(
        """grep '/var/cache/apt/archives' /var/log/unattended-upgrades/unattended-upgrades.log | grep -v check_conffile | grep AcquireItem"""
    )().splitlines():
        match = TARGET_MATCHER.search(line)
        if match:
            filename = match.group('filename')
            base = os.path.basename(filename)
            package = base.split('_')[0]
            if package in unique:
                continue
            unique.add(package)
            if excluded:
                if filter_exclusions( package, excluded ):
                    result.append( filename )
            else:
                result.append( filename )
    return result

def metadata_main():
    options = get_metadata_options().parse_args()
    structure = metadata_structure( options.filenames )
    if options.merge:
        new_set = dict([(s['package'],s) for s in structure['metadata']])
        final_set = []
        previous = json.loads(open(options.merge).read())
        for item in previous['metadata']:
            if item['package'] in new_set:
                log.info("Updating record for %s",item['package'])
                item = new_set.pop(item['package'])
            final_set.append( item )
        final_set.extend( new_set.values())
        structure['metadata'] = final_set
        previous.update(structure)
        structure = previous
    print(json.dumps(structure,indent=2))
    
def metadata_structure( filenames ):
    metadata_list = []
    for package in filenames:
        metadata_list.append( metadata(package))
    return {
        'metadata': metadata_list, 
        'ts': time.time(), 
    }

def metadata( filename ):
    command = ['dpkg', '--info', filename]
    
    output = nbio.Process(command)()
    for record in parse_dpkg_info(output).values():
        record['filename'] = os.path.basename(filename)
        return record 
    raise RuntimeError("Unable to extract metadata for %r"%(filename, ))

def pack_zip(target='os-upgrades.deb.tar', excluded=None):
    log.info("Packing into target: %r", target)
    t = tarfile.TarFile(target, mode='w')
    try:
        packages = get_components(excluded=excluded)
        for package in packages:
            log.info("Packing: %s", package)
            t.add( package, os.path.basename(package) )
        structure = metadata_structure( packages )
        try:
            twrite.twrite( target + '.json', json.dumps(structure))
            t.add( target+'.json',  'metadata.json' )
        finally:
            os.remove( target + '.json' )
    finally:
        t.close()

def main():
    logging.basicConfig(level=logging.INFO)
    options = get_options().parse_args()
    if options.recalculate:
        recalculate()
    if options.list:
        for package in get_components(excluded=options.excluded):
            base = os.path.basename(package)
            package, version = base.split('_', 1)
            print( '%s\t%s'%(package, version))
        return
    pack_zip( target=options.target,  excluded=options.excluded )
    return 0

if __name__ == "__main__":
    main()



"""Transactionally write a file to disk"""
import os,logging
from ._shims import bytes, unicode
log = logging.getLogger( __name__ )

def twrite( filename, content ):
    """Write content to filename, creating directory if it does not exist
    
    Notes:
        Is not race-condition safe wrt multiple writers, as it uses 
        the same filename for the temp file in each write.
        
        Transactionality is Linux specific (os.rename guarantee)
    """
    directory = os.path.dirname( filename )
    if directory and not os.path.exists( directory ):
        log.debug( 'Creating directory: %s', directory )
        os.makedirs( directory )
    tmp = os.path.join( filename+'~' )
    fh = open( tmp, 'wb' )
    if isinstance( content, bytes ):
        fh.write( content )
    elif isinstance( content, unicode ):
        fh.write( content.encode( 'utf-8' ))
    else:
        for item in content:
            fh.write( item )
    fh.close()
    log.debug( 'Writing file: %s', filename )
    try:
        os.rename( tmp, filename )
    except OSError as err:
        if err.errno == 2: # no such directory
            pass
        else:
            raise 
    



"""Provide a cron lock for preventing cron-bombs and the like"""
import os, logging, fcntl, errno, signal, tempfile, time, threading
log = logging.getLogger( __name__ )

def time_function( marker=None ):
    from functools import wraps
    def wrapper( function ):
        if marker is None:
            final_marker = '%s.%s'%(function.__module__,function.__name__)
        else:
            final_marker = marker
        @wraps( function )
        def timer( *args, **named ):
            start = time.time()
            try:
                holder = open( '/var/firmware/run/request-localhost-8111.lock' ).read()
            except Exception:
                holder = None
            try:
                return function( *args, **named )
            finally:
                log.info("%s: %0.1fs by %s", final_marker, time.time() - start, holder )
        return timer
    return wrapper

class Busy( IOError ):
    """Raised if the lock is held by another cronlock"""
class Timeout( RuntimeError ):
    """Raised if the timeout signal triggers"""
class AlreadyHeld( Busy ):
    """Raised if we attempt to lock multiple times from the same process"""
    
class Flock( object ):
    """A context manager that just flocks a file"""
    file = None
    def __init__( self, filename, blocking=False ):
        self.filename = filename 
        self.blocking = blocking
    def __str__( self ):
        return '%s( %r )'%( self.__class__.__name__, self.filename )
    def __enter__( self ):
        if self.file:
            raise Busy( """Attempted to lock the %s twice"""%( self ))
        fh = open( self.filename, 'ab+' )
        for i in range(2): # limited number of retries
            try:
                flags = fcntl.LOCK_EX
                if not self.blocking:
                    flags |= fcntl.LOCK_NB
                fcntl.flock( fh, flags ) # can raise IOError 
            except IOError as err:
                if err.errno == errno.EWOULDBLOCK:
                    other_pid = None
                    if os.path.exists(self.filename):
                        try:
                            other_pid = int(open(self.filename).readline())
                        except Exception:
                            pass
                        else:
                            from fussy import processcontrol 
                            if os.getpid() == other_pid:
                                raise AlreadyHeld(
                                    "Multiple attempts to lock the same file from the same process %s"%(self.filename)
                                )
                            if not processcontrol.alive( other_pid ):
                                log.info(
                                    "PID#%s held the lock on %s, but does not appear to be alive, retrying",
                                    other_pid,
                                    self.filename,
                                )
                                continue
                            
                    err = Busy( *(err.args + ('held by PID#%s'%(other_pid,), )) )
                    err.errno = errno.EWOULDBLOCK
                raise err
        fh.seek( 0 ) # rewind to start...
        fh.write( (u'%s\n'%(os.getpid())).encode('utf-8'))
        fh.flush()
        self.file = fh
        return fh
    def __exit__( self, *args ):
        try:
            fh = self.__dict__.pop( 'file' )
        except KeyError:
            log.warn("No 'file' member in lock, somehow exiting before entry completed?")
        else:
            try:
                os.remove(self.filename)
            except (OSError, IOError):
                log.warn('File %s removed while locked', self.filename)
            fcntl.flock( fh, fcntl.LOCK_UN )
            fh.close()
    def write( self, *args ):
        with self:
            return self._write(*args)
    def _write(self,*args):
        self.file.seek(0)
        self.file.write( *args )
        self.file.flush()
    def read( self, *args ):
        with self:
            return self._read(*args)
    def _read(self, *args):
        return self.file.read( *args )

class Lock( object ):
    """A Context manager that provides cron-style locking"""
    def __init__( self, lockfile ):
        """Create a lock file 
        
        name -- used to construct the lock-file name 
        directory -- directory in which to construct the lock-file 
        """
        self.flock = Flock( lockfile )
    @property 
    def pid( self ):
        return os.getpid()
    def __enter__( self ):
        self.flock.__enter__()
    def __exit__( self, *args ):
        self.flock.__exit__(*args)
    def on_timeout( self, *args, **named ):
        raise Timeout( "Maximum run-time exceeded, aborting" )
    def set_timeout( self, duration ):
        """Set a signal to fire after duration and raise an error"""
        signal.signal( signal.SIGALRM, self.on_timeout )
        signal.alarm( int(duration) )

class ThreadLock(object):
    """Simple context manager for a threading RLock"""
    def __init__(self, lock_wait=5):
        self._lock = threading.RLock()
        self.lock_wait = lock_wait
    def __enter__(self):
        start = time.time()
        abort = start + self.lock_wait
        while time.time() < abort:
            if not self._lock.acquire(blocking=False):
                time.sleep(max(
                    (
                        (abort-time.time())/100., 
                        0.00001
                    ), 
                ))
            else:
                #log.debug("Acquired thread lock after: %0.3fs", time.time()-start)
                return self
        raise Busy('Timeout waiting for thread lock')
    def __exit__(self, *args):
        self._lock.release()

class BlockingLock( Lock ):
    """Lock class that blocks for up to lock_wait seconds waiting for other process to exit
    
    Also uses ThreadLock to prevent multiple threads from 
    this process from running simultaneously...
    """
    def __init__( self, lockfile, lock_wait=5 ):
        super( BlockingLock, self ).__init__( lockfile )
        self._threadlock = ThreadLock()
        self.lock_wait = lock_wait
    def __enter__( self ):
        start = time.time()
        abort = start + self.lock_wait
        while time.time() < abort:
            try:
                result = super( BlockingLock, self ).__enter__()
                delta = time.time() - start 
                if delta > .1:
                    log.info("Acquired after %0.1fs",delta)
                return result 
            except Busy:
                max_wait = .01
                duration = max(
                    (
                        min(( 
                            (abort-time.time())/100.,
                            max_wait,
                        )), 
                        0.00001
                    ), 
                )
                time.sleep(duration)
        log.debug("Timeout waiting")
        raise Busy( "Lock on %s timed out"%( self.flock.filename ))

def with_lock( name, directory=None, timeout=None ):
    """Decorator that runs a function with Lock instance acquired
    
    * name -- basename of the file to create 
    * directory -- if specified, the directory in which to store files, 
        defaults to tempfile.gettempdir()
    * timeout -- if specified, the number of seconds to allow before raising 
        a Timeout error
    """
    if directory is None:
        directory = tempfile.gettempdir()
    filename = os.path.join( directory, name )
    lock = Lock( filename )
    def wrap_func( function ):
        """Wraps the function execution with our lock"""
        def wrapped_with_lock( *args, **named ):
            with lock:
                if timeout:
                    lock.set_timeout( timeout )
                return function( *args, **named )
        wrapped_with_lock.__name__ = function.__name__
        wrapped_with_lock.__doc__ = function.__doc__
        wrapped_with_lock.__dict__ = function.__dict__
        wrapped_with_lock.lock = lock 
        return wrapped_with_lock
    return wrap_func



"""Tail protocol for files using Twisted non-blocking IO"""
import os, sys, errno
from twisted.internet import fdesc, reactor
from twisted.python import log

from twisted.internet import inotify
from twisted.python import filepath

class FileTail( object ):
    def __init__(self,filename ):
        if hasattr(filename,'path'):
            filename = filename.path
        self.filename = filename
        self.buffer = ''
    def start_watching(self):
        # TODO: allow for handling log rotation etc.
        # TODO: use the overall directory inotify instead,
        # and use *that* to do log-rotation checks...
        self.notifier = inotify.INotify()
        self.notifier.startReading()
        self.notifier.watch(
            filepath.FilePath(self.filename), 
            mask = inotify.IN_MODIFY,
            callbacks=[self.on_file_change],
        )
    _fd = None
    @property
    def fd(self):
        if self._fd is None:
            self._fd = os.open(self.filename,os.O_RDONLY )
            fdesc.setNonBlocking( self._fd )
        return self._fd
    def on_file_change(self,ignored, filepath, mask):
        if mask & inotify.IN_MODIFY:
            self.read_from_file()
    def read_from_file(self):
        fd = self.fd
        while True:
            # TODO: really, we should do a non-blocking producer 
            # using an abstract file handle, but as that turns 
            # out not to usefully restrict us on EOF it's kinda 
            # loopy...
            # For the simple "tail a file to web browser" case 
            # this should work...
            try:
                output = os.read( fd, 8192 )
            except  (OSError, IOError) as err:
                if err.args[0] in (errno.EAGAIN, errno.EINTR):
                    pass
                elif err.args[0] in errno.EWOULDBLOCK:
                    # can't read any more
                    break
                else:
                    log.msg("Error reading from %s"%(self.name,))
                    self._fd = None
            else:
                if output:
                    self.buffer += output
                    log.msg('Got output: %s'%(len(self.buffer)))
                else:
                    # reached end of file...
                    break
        if self.buffer:
            lines = self.buffer.splitlines()
            if not self.buffer.endswith( '\n' ):
                self.buffer = lines.pop()
            else:
                self.buffer = ''
            return lines
        return []

def test_consumer( lines ):
    for line in lines:
        log.msg('line: %r'%(line,))

def main():
    log.startLogging(sys.stderr)
    FileTail( sys.argv[1], test_consumer ).start_watching()
    reactor.run()

if __name__ == "__main__":
    main()
    



"""Load firmware configuration so that we can find out firmware metadata

Expected keys:

    [fussy]
    product=sample
    keyring=/etc/fussy/keys
    extra_etc=/etc/%(product)s
    var_directory=/var/%(sample)s
    firmware_directory=/opt/%(sample)s
"""
import glob
try:
    import ConfigParser as configparser 
except ImportError:
    import configparser

options = None
def loadoptions(product='fussy'):
    global options
    if options is None:
        options = configparser.SafeConfigParser( )
        options.read( sorted(glob.glob( '/etc/fussy/*.conf' )))
        try:
            product = options.get( 'fussy','product' )
        except (configparser.NoSectionError,configparser.NoOptionError) as err:
            pass 
        else:
            options.read( sorted(glob.glob( '/etc/%(product)s/*.conf'%locals() ) ))
    return options 

def getstring( section, key, default=None ):
    if not options:
        raise RuntimeError( "Need to load options first")
    try:
        return options.get( section, key )
    except (configparser.NoSectionError,configparser.NoOptionError):
        return default 
def getboolean( section, key, default=None ):
    if not options:
        raise RuntimeError( "Need to load options first")
    try:
        return options.getboolean( section, key )
    except (configparser.NoSectionError,configparser.NoOptionError):
        return default 



"""Wraps subprocess with pipe semantics and generator based multiplexing

pipe = open( somefile, 'rb' ) | nbio.Process( ['grep','blue'] ) | nbio.Process( ['wc','-l'])

"""
import os, select, fcntl, time, sys, logging, subprocess, errno, collections
log = logging.getLogger( __name__ )
from ._shims import bytes, unicode, file_base, next, long

DEFAULT_READ_BLOCK = 1024*64

def set_non_blocking( fh ):
    """Set this file handle to non-blocking mode"""
    current = fcntl.fcntl( fh, fcntl.F_GETFL )
    fcntl.fcntl( fh, fcntl.F_SETFL, current | os.O_NONBLOCK )
    return fh

class _DidNothing( object ):
    """Returned when a subprocess didn't do anything"""
    def __bool__( self ):
        return False
    __nonzero__ = __bool__
DID_NOTHING = _DidNothing()

class NBIOError( RuntimeError ):
    """Base class for nbio errors"""
class ProcessError( NBIOError ):
    """Called process returned an error code
    
    Attributes:
    
        process -- the Process (if applicable) which raised the error
        exitcode -- exit code from the process if available
    """
    process = None
    exitcode = None
class Timeout( NBIOError ):
    """Called process did not complete within specified timeout
    
    Attributes:
    
        process -- the Process (if applicable) which raised the error
    
    """
    process = None

def pause( duration ):
    """Allow tests to override sleeping using globalsub"""
    time.sleep( duration )

class Pipe( object ):
    """Pipe of N processes which all need to process data in parallel"""
    pause_on_silence = 0.01
    started = False
    def __init__( self, *processes ):
        self.processes = []
        if processes:
            self.append( self.get_component(processes[0]) )
            for process in processes[1:]:
                self.__or__( process )
    def __repr__( self ):
        return '%s => %s'%( 
            self.__class__.__name__, ' | '.join( [str(p) for p in self.processes]) 
        )
    def __getitem__(self, index):
        """Retrieve a particular item in the pipe"""
        return self.processes[index]
    def __len__( self ):
        """Return the number of items in this pipe"""
        return len( self.processes )
        
    def __iter__( self, timeout=None ):
        """Iterate over the processes in the pipe
        
        If the stdout/stderr of the processes is not captured, then we will 
        yield the results in whatever chunk-size is yielded from the individual 
        processes.
        
        If all of the processes yield DID_NOTHING in a particular cycle, then the 
        pipe will do a pause() for self.pause_on_silence (normally passed into the 
        __call__) before the next iteration.
        """
        if timeout:
            until = time.time() + timeout
        try:
            self.started = True
            iterables = [ iter(process) for process in self.processes ]
            crashed = False
            while iterables and not crashed:
                exhausted = []
                something_happened = 0
                for iterable in iterables:
                    try:
                        child_result = next(iterable)
                        if child_result is DID_NOTHING:
                            # we did nothing during this iteration
                            pass
                        elif child_result:
                            something_happened += 1
                            yield child_result
                    except StopIteration:
                        exhausted.append( iterable )
                    except (KeyboardInterrupt,Exception):
                        # Have the *first* item that failed raise the error, 
                        # and only if they don't, raise this error...
                        for item in self.processes:
                            item.check_exit()
                        raise
                if not something_happened:
                    for item in self.processes:
                        if hasattr( item, 'poll' ):
                            item.poll()
                # check for a process/pipe/file failure...
                for done in exhausted:
                    while done in iterables:
                        iterables.remove( done )
                if not something_happened:
                    try:
                        pause( self.pause_on_silence )
                    except (KeyboardInterrupt,Exception) as err:
                        new_err = ProcessError( 
                            "Exception while paused: %s"%(err,)
                        )
                        new_err.process = self 
                        raise new_err
                if timeout:
                    if time.time() > until:
                        raise Timeout( self, timeout )
        finally:
            self.kill()
    def __call__( self, pause_on_silence=0.01, timeout=None ):
        """Iterate over this pipeline, returning combined results as a string
        """
        result = []
        self.pause_on_silence = pause_on_silence
        try:
            for item in self.__iter__( timeout=timeout ):
                if item:
                    result.append( item )
            return b"".join( result )
        except (KeyboardInterrupt,Exception) as err:
            err.output = result
            raise
    def append( self, process ):
        """Add the given PipeComponent to this pipe (note: does not connect stdin/stdout)"""
        assert isinstance( process, PipeComponent ), process
        self.processes.append( process )
    def prepend( self, process ):
        """Add the given PipeComponent to this pipe (note: does not connect stdin/stdout)"""
        assert isinstance( process, PipeComponent ), process
        self.processes.insert( 0, process )
    @property
    def first( self ):
        """Retrieves the first item in the pipe"""
        if self.processes:
            return self.processes[0]
        raise IndexError( "No processes yet in this pipeline" )
    @property
    def last( self ):
        """Retrieves the last item in the pipe"""
        if self.processes:
            return self.processes[-1]
        raise IndexError( "No processes yet in this pipeline" )

    def __or__( self, other ):
        """Pipe our output into a process, callable or list"""
        other = self.get_component( other )
        last = self.last 
        # special cases for piping process-to-process 
        both_processes = isinstance(last,Process) and isinstance( other, Process )
        both_processes = False
        if both_processes and not other.started and not last.started:
            log.info( 'Direct pipe %s -> %s', last, other )
            last.blocking_stdout()
            other.direct_pipe( last.stdout, 'stdin' )
        else:
            last.iterables.append(
                other.iter_write( self.last.iter_read() )
            )
        self.append( other )
        return self
    def __ror__( self, other ):
        """Pipe output of other into our first item"""
        other = self.get_component( other )
        self.first.iterables.append(
            self.first.iter_write( 
                other.iter_read()
            )
        )
        self.prepend( other )
        return self

    def __gt__( self, other ):
        """Pipe our output into a file
        """
        if isinstance( other, (bytes,unicode)):
            if other not in ('','-'):
                return self.__or__( open(other,'wb') )
            else:
                return self.__or__( other )
        else:
            return self.__or__( other )
    def __lt__( self, other ):
        """Pipe input from a named file"""
        if isinstance( other, (bytes,unicode)):
            if other not in ('','-'):
                return self.__ror__( open(other,'rb') )
            else:
                return self.__ror__( other )
        else:
            return self.__ror__( other )
    def get_component( self, other ):
        """Given a python object other, create a PipeComponent for it
        
        The purpose of this method is to allow for fairly "natural" 
        descriptions of tasks.  You can pipe to or from files, 
        to or from the string '-' (stdin/stdout), to the 
        string '' (collect stdout), or from a regular string (which is 
        treated as input).  You can pipe iterables into a pipe,
        you can pipe the result of pipes into callables.
        """
        if isinstance( other, PipeComponent ):
            pass 
        elif isinstance( other, file_base ):
            other = FileComponent( other )
        elif isinstance( other, (bytes,unicode)):
            if other == '-':
                other = FileComponent( sys.stdout, sys.stdin )
            elif other == '':
                other = IterComponent( )
            else:
                other = IterComponent( [other] )
        elif hasattr( other, '__iter__'):
            other = IterComponent( other )
        elif isinstance(callable, collections.Callable):
            other = FunctionComponent( other )
        if not isinstance( other, PipeComponent ):
            raise TypeError( """Require a PipeComponent-compatible object, got: %s"""%(other,) )
        return other 
    def kill( self ):
        for process in self.processes:
            if hasattr( process, 'kill' ):
                try:
                    process.kill()
                except (KeyboardInterrupt,Exception):
                    log.warn(
                        "Unable to kill process: %s", process,
                    )
    

class PipeComponent( object ):
    live = True
    def __init__( self ):
        self.iterables = []
    def __iter__( self ):
        iterables = [ iter(x) for x in self.iterables ]
        while iterables and self.live:
            something_happened = 0
            exhausted = []
            for iterable in iterables:
                try:
                    child_result = next(iterable)
                    if child_result is DID_NOTHING:
                        pass
                    elif child_result:
                        something_happened += 1
                        yield child_result
                except StopIteration as err:
                    exhausted.append( iterable )
                except (KeyboardInterrupt,Exception) as err:
                    err.args += (self,)
                    raise 
            for done in exhausted:
                while done in iterables:
                    iterables.remove( done )
            if not something_happened:
                yield DID_NOTHING
        self.check_exit()
    def check_exit( self ):
        pass
    def iter_read( self ):
        """Iterate reading from stdout"""
        raise TypeError( "Do not have an iter read for %s"%(self.__class__, ))
    def iter_write( self, source ):
        raise TypeError( "Do not have an iter write for %s"%(self.__class__, ))
    

class FileComponent( PipeComponent ):
    def __init__( self, filein,fileout=None ):
        self.stdin = filein
        self.stdout = fileout or filein 
        super( FileComponent, self ).__init__()
    def iter_read( self ):
        return reader( self.stdout, exit_on_finish=True )
    def iter_write( self, source ):
        return writeiter(
            source,
            self.stdin,
        )
    def __repr__( self ):
        if self.stdin != self.stdout:
            return '%s/%s'%( self.stdout, self.stdin )
        else:
            return 'file( %r, %r )'%( self.stdin.name, self.stdin.mode )
    
class Process( PipeComponent ):
    """A particular process in a Pipe
    
    Processes are the most common entry point when using nbio, you 
    create processes and pipe data into or out of them as appropriate
    to create Pipes.
    
    Under the covers the Process runs subprocess.Popen, and it accepts most of 
    the fields subprocess.Popen does.  By default it captures stdout and pipes 
    data into stdin.  If nothing is connected to stdin then stdin is closed on 
    the first iteration of the pipe.  If nothing is connected to stdout or 
    stderr (if stderr is captured) then the results will be returned to the 
    caller joined together with ''
    
    The implication is that if you do not want to store all of the results 
    in RAM, you need to "sink" the results into a process or file, or *not*
    capture the results (pass False for stdout or stderr).
    """
    stdin_needed = False
    stdout_needed = False 
    stderr_needed = False
    STDOUT = -1
    by_line = False
    def __init__( self, command, stderr=False, stdout=True, stdin = True,**named ):
        """Initialize the Process 
        
        command -- subprocess.Popen command string or list
                   if a string, and "shell" is not explicitly set, 
                   then will set "shell=True"
                   
        stdin -- whether to provide stdin writing 
        
        stdout -- whether to capture stdout 
        
        stderr -- whether to capture stderr, if -1, then combine stdout and stderr 
        
        good_exit -- if provided, iterable which provides the set of good exit codes 
                     which will not raise ProcessError when encountered
                     
        by_line -- if provided, will cause the output to be line-buffered so that 
                   only full lines will be reported, the '\\n' character will be used to 
                   split the output, so there will be no '\\n' character at the end of each 
                   line.
                   
        named -- passed to the subprocess.Popen() command 
        
        """
        if isinstance( command, (bytes,unicode)):
            if not 'shell' in named:
                named['shell'] = True 
        self.command = command 
        if 'good_exit' in named:
            self.good_exit = named.pop( 'good_exit' )
        else:
            self.good_exit = [0]
        if 'by_line' in named:
            self.by_line = named.pop( 'by_line' )
        named['stdin'],named['stdout'],named['stderr'] = stdin,stdout,stderr
        self.named_args = named
        try:
            self.no_descendent_kill = named.pop( 'no_descendent_kill' )
        except KeyError:
            self.no_descendent_kill = False
        super( Process, self ).__init__( )
    def __unicode__( self ):
        return u'%s( %r )'%( self.__class__.__name__, self.command )
    __repr__ = __unicode__
    
    @property
    def stderr( self ):
        return self.pipe.stderr
    @stderr.setter
    def stderr( self, new ):
        if self._pipe:
            raise ProcessError( 'Cannot set stderr after pipe start' )
        self.named_args['stderr'] = new 
    @property
    def stdout( self ):
        return self.pipe.stdout
    @stdout.setter
    def stdout( self, new ):
        if self._pipe:
            raise ProcessError( 'Cannot set stdout after pipe start' )
        self.named_args['stdout'] = new 
    @property
    def stdin( self ):
        return self.pipe.stdin
    @stdin.setter
    def stdin( self, new ):
        if self._pipe:
            raise ProcessError( 'Cannot set stdin after pipe start' )
        self.named_args['stdin'] = new 

    use_blocking_stdout = False
    def blocking_stdout( self ):
        self.use_blocking_stdout = True
    def direct_pipe( self, file, key='stdout' ):
        """Directly pipe stdout/stderr/stdin to/from the file"""
        setattr( self, key, file )
        setattr( self, '%s_wanted'%(key,), True )
        self.iterables.append( self.iter_wait() )
        return file
    
    @property 
    def pid( self ):
        return self.pipe.pid
    @property 
    def started( self ):
        return self._pipe is not None
    _pipe = None
    @property 
    def pipe( self ):
        if self._pipe is None:
            self._pipe = self._start_pipe( 
                **self.named_args 
            )
        return self._pipe
    def start(self):
        """Trigger to start the process immediately"""
        # yes, side-effects are ick...
        self.pipe
        return self
    
    def _start_pipe( self, **named ):
        """Start the captive process (internal operation)"""
        for key in ('stdin','stdout','stderr'):
            value = named.get(key)
            if value == -1:
                value = subprocess.STDOUT
            elif hasattr( value, 'fileno' ):
                log.info( 'Using existing file: %s', value )
                value = value
            else:
                value = [None,subprocess.PIPE][bool(value)]
            named[key] = value 
        
        pipe = subprocess.Popen(
            self.command, 
            **named
        )
        for key in ('stdin','stdout','stderr'):
            if not hasattr(named[key],'fileno'):
                if key == 'stdout' and self.use_blocking_stdout:
                    log.info( 'Using blocking %s', key )
                    continue 
                    #pass
                fh = getattr( pipe, key, None )
                if fh is not None:
                    set_non_blocking( fh )
            else:
                log.info( 'Using blocking %s', key )
        return pipe
    def __iter__( self ):
        """Iterate over the results of the process (normally done by the Pipe)"""
        if not self.stdout_needed:
            # nothing has been hooked to stdout
            if self.stdout:
                # but stdout was captured, so we need to consume it to prevent deadlocks
                self.iterables.append(
                    self.iter_read()
                )
        if not self.stderr_needed:
            # nothing has been hooked to stderr
            if self.stderr:
                self.iterables.append(
                    self.iter_read_err( )
                )
        if not self.stdin_needed:
            # nothing is being sent in, so input is finished...
            close( self.stdin )
        return super( Process, self ).__iter__()
    
    def __or__( self, other ):
        """Pipe our output into a process, callable or list
        
        pipe = Pipe( Process( 'cat test.txt' ) | Process( 'grep blue' ) | [] )
        pipe()
        """
        self.stdout_needed = True 
        if isinstance( other, Pipe ):
            # Pipe our output into the pipe, add ourselves to the start of pipe...
            return other.__nor__( self )
        else:
            pipe = Pipe( self )
            return pipe.__or__( other )
        
    def __ror__( self, other ):
        """Pipe other into self"""
        self.stdin_needed = True 
        if isinstance( other, Pipe ):
            return other.__or__( self )
        else:
            pipe = Pipe( self )
            return pipe.__ror__( other )
    def __gt__( self, other ):
        """Pipe our output into a filename"""
        pipe = Pipe( self )
        return pipe.__gt__( other )
    def __lt__( self, other ):
        """Pipe our input from a filename"""
        pipe = Pipe( self )
        return pipe.__lt__( other )
    def __call__( self, *args, **named ):
        """Create a Pipe and run it with just this item as its children"""
        pipe = Pipe( self )
        return pipe( *args, **named )
        
    def poll( self ):
        return self.pipe.poll()
    
    def check_exit( self ):
        """Check our exit code"""
        if self.pipe.returncode is None:
            exitcode = self.pipe.poll()
        else:
            exitcode = self.pipe.returncode
        if exitcode is None or exitcode < 0:
            try:
                if not self.no_descendent_kill:
                    from fussy.processcontrol import kill_children
                    kill_children( self.pipe.pid )
                else:
                    self.kill()
                    
            except OSError as err:
                if err.errno == errno.ESRCH:
                    # we exited...
                    pass 
                else:
                    raise
            exitcode = self.pipe.poll()
        if exitcode is None:
            log.error(
                "Process %s appears not to have exited properly", self 
            )
        elif exitcode not in self.good_exit:
            err = ProcessError( 
                "Process %s returned error code %s"%( self.command, exitcode, ),
            )
            err.process = self 
            err.exitcode = exitcode
            raise err

    def iter_read( self ):
        """Create the thing which iterates our read operation"""
        self.stdout_needed = True
        output = reader( self.stdout )
        if self.by_line:
            output = by_line( output )
        return output
    def iter_read_err( self ):
        self.stderr_needed = True 
        output = reader( self.stderr )
        if self.by_line:
            output = by_line( output )
        return output
    def iter_write( self, source ):
        """Create a thing which will read from source and write to us"""
        self.stdin_needed = True
        return writeiter(
            source,
            self.stdin,
        )
    def iter_wait( self ):
        while self.pipe.poll() == None:
            yield DID_NOTHING
    def kill( self ):
        """Kill our underlying subprocess.Popen"""
        try:
            return self.pipe.kill( )
        except OSError:
            # we've already died/closed...
            pass 
    
class FunctionComponent( PipeComponent ):
    def __init__( self, function ):
        self.function = function 
        super( FunctionComponent, self ).__init__()
    def iter_read( self ):
        yield self.function()
    def iter_write( self, source ):
        return caller( source, self.function )
    def __str__( self ):
        return str( self.function )

class IterComponent( PipeComponent ):
    def __init__( self, iterable=None  ):
        self.iterable = iterable 
        super( IterComponent, self ).__init__()
    def iter_read( self ):
        return self.iterable
    def iter_write( self, other ):
        return other 
    def __repr__( self ):
        if self.iterable:
            return str( self.iterable )
        else:
            return '<str>'

def collector( iterable, target ):
    return caller( iterable, target.append )
def caller( iterable, target ):
    for item in iter(iterable):
        if item:
            target( item )
        yield item

#def pipeline( *commands ):
#    pipe = []
#    previous_stdout = subprocess.PIPE
#    for command in commands:
#        pipe.append( subprocess.Popen( command, stdin=previous_stdout, stdout=subprocess.PIPE ))
#        previous_stdout = pipe[-1].stdout 
#    return pipe

def writeiter( iterator, fh ):
    """Write content from iterator to fh
    
    To write a file from a read file:
    
    .. code-block:: python
    
        writeiter(
            reader( open( filename )),
            fh 
        )
    
    To write a request.response object into a tar pipe iteratively:
    
    .. code-block:: python
    
        writeiter( 
            response.iter_content( DEFAULT_READ_BLOCK, decode_unicode=False ),
            pipe 
        )
    """
    total_written = 0
    assert hasattr( iterator, '__iter__'), iterator
    for content in iterator:
        if content and isinstance( content, (bytes,unicode)):
            for block_written in writer( content, fh ):
                if block_written is DID_NOTHING:
                    yield block_written
                else:
                    total_written += block_written
                    yield None
        else:
            yield DID_NOTHING
    #log.debug( 'Closing %s after %s bytes', fh, total_written )
    close( fh )

def writer( content,fh, encoding='utf8', block_size=4096 ):
    """Continue writing content (string) to fh until content is consumed
    
    Used by writeiter to writing individual bits of content to the fh
    """
    if hasattr( fh, 'raw' ):
        fh = fh.raw
    fp = fileno( fh )
    if isinstance( content, unicode ):
        content = content.encode( encoding )
    assert isinstance( content, bytes ), """Can only write byte values: %r"""%(content)
    for block in string_blocks( content, block_size=block_size ):
        while block:
            written = 0
            try:
                written = os.write( fp, block )
            except OSError as err:
                if err.errno == errno.EWOULDBLOCK:
                    yield DID_NOTHING 
                else:
                    log.warn( "OS Error writing: %s", err )
                    return
            except (IOError,ValueError) as err:
                log.warn( "Error while writing: %s", err )
                return
            if not written:
                #log.debug( 'Unable to write' )
                yield DID_NOTHING
            else:
                #log.debug( 'Wrote: %s', written )
                block = block[written:]
                yield written
        
def reader( fh, blocksize = DEFAULT_READ_BLOCK, exit_on_finish=False ):
    """Produce content blocks from fh without blocking
    """
    if hasattr( fh, 'raw' ):
        fh = fh.raw
    try:
        fd = fileno( fh )
    except ValueError as err:
        return 
    while not fh.closed:
        try:
            rr,rw,re = select.select([fh],[],[],0)
            if rr:
                result = os.read(fd, blocksize )
                if result == b'':
                    if exit_on_finish:
                        close(fh)
                        return
                    break
            else:
                result = DID_NOTHING
        except ValueError as err:
            break 
        except (IOError,OSError) as err:
            if err.args[0] in ( errno.EWOULDBLOCK,):
                if fh.closed:
                    break
                yield DID_NOTHING
            else:
                raise 
        else:
            yield result 
    close( fh )

def fileno( fh ):
    "Determine the fileno for the file-like thing"
    if hasattr( fh, 'raw' ):
        fh = fh.raw
    if hasattr( fh, 'fileno' ):
        return fh.fileno()
    return fh
def close( fh ):
    """Close the file/socket/closable thing"""
    if fh is sys.stdout or fh is sys.stderr:
        return
    if hasattr( fh, 'close'):
        errcode = fh.close()
    else:
        fn = fileno( fh )
        if fn and isinstance( fn, (int,long)):
            errcode = os.close( fileno(fh) )
        else:
            errcode = 0
    if errcode:
        raise RuntimeError( "Child process returned error code", errcode )
def by_line( iterable ):
    """Buffer iterable yielding individual lines"""
    buffer = b""
    for item in iterable:
        if not item:
            yield item 
        else:
            buffer += item 
            if b'\n' not in item:
                # We did something, but we're not ready to yield a real value...
                yield None
            else:
                while b'\n' in buffer:
                    line, buffer = buffer.split(b'\n',1)
                    yield line 
    if buffer:
        yield buffer 

def string_blocks( content, block_size=4096 ):
    """Yield blocks from content only producing one block at a time"""
    current = 0
    length = len(content)
    if len(content) < block_size:
        # no point copying it...
        yield content 
    else:
        while current < length:
            new = current + block_size
            yield content[current:new]
            current = new



"""Pure-python inotify interface

This provides the same basic API as the inotifyx package,
but does not require a binary installation. Notable changes
from inotifyx are that the init() call used is *always* 
the non-blocking-flagged init1 call (i.e. we always use 
non-blocking IO) and that we make events' __repr__ show the 
event flags by default.

In addition, a trivial OO facade is provided that tracks the 
watches added to an Inotify instance and uses that to provide
a bit of metadata, such as the final path-name for InotifyEvent
instances.

Example usage for the function-based API:

.. doctest::

    >>> import os,shutil,tempfile
    >>> temp = tempfile.mkdtemp(prefix='inotify-',suffix='-test')
    >>> fd = init()
    >>> final_dir = os.path.join(temp,'path','to','directory')
    >>> os.makedirs(final_dir)
    >>> wd = add_watch(fd,final_dir,IN_ALL_EVENTS)
    >>> list(get_events(fd))
    []
    >>> fh = open(os.path.join(final_dir,'test.txt'),'w')
    >>> x = fh.write('oo')
    >>> fh.close()
    >>> for event in get_events(fd):
    ...    print(event) # doctest: +ELLIPSIS
    InotifyEvent(...,...,IN_CREATE,0,'test.txt')
    InotifyEvent(...,...,IN_OPEN,0,'test.txt')
    InotifyEvent(...,...,IN_MODIFY,0,'test.txt')
    InotifyEvent(...,...,IN_CLOSE_WRITE,0,'test.txt')
    >>> list(get_events(fd,.001)) # block for up to 1ms
    []
    >>> rm_watch(fd,wd)
    >>> add_watch(fd,os.path.join(temp,'moo','roo')) # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
      File "/usr/lib/python2.7/doctest.py", line 1315, in __run
        compileflags, 1) in test.globs
      File "<doctest fussy.inotify[13]>", line 1, in <module>
        add_watch(fd,os.path.join(temp,'moo','roo')) # doctest: +IGNORE_EXCEPTION_DETAIL
      File "/mnt/homevar/var/pylive/fussy/fussy/inotify.py", line 115, in wrapped
        raise InotifyError( err, errno.errorcode.get(err, 'Unknown error') )
    InotifyError: [Errno 0] Unknown error
    >>> shutil.rmtree(temp)

"""
import ctypes, os, errno, sys
from functools import wraps
from ._shims import unicode

def as_8_bit(s):
    if isinstance(s, unicode):
        return s.encode('utf-8')
    return s
def clean_repr(s):
    base = repr(s)
    while base.startswith('u'):
        base = base[1:]
    return base
NULL_BYTE = as_8_bit('\000')

class InotifyError(OSError):
    """Class for errors raised by Inotify"""

IN_ACCESS=0x00000001
IN_MODIFY=0x00000002
IN_ATTRIB=0x00000004
IN_CLOSE_WRITE=0x00000008
IN_CLOSE_NOWRITE=0x00000010
IN_OPEN=0x00000020
IN_MOVED_FROM=0x00000040
IN_MOVED_TO=0x00000080
IN_CREATE=0x00000100
IN_DELETE=0x00000200
IN_DELETE_SELF=0x00000400
IN_MOVE_SELF=0x00000800
IN_UNMOUNT=0x00002000
IN_Q_OVERFLOW=0x00004000
IN_IGNORED=0x00008000
IN_ONLYDIR=0x01000000
IN_DONT_FOLLOW=0x02000000
IN_EXCL_UNLINK=0x04000000
IN_MASK_ADD=0x20000000
IN_ISDIR=0x40000000
IN_ONESHOT=0x80000000
EVENT_NAMES = [
    'IN_ACCESS', 'IN_MODIFY', 
    'IN_ATTRIB', 'IN_CLOSE_WRITE', 
    'IN_CLOSE_NOWRITE', 'IN_OPEN', 
    'IN_MOVED_FROM', 'IN_MOVED_TO', 
    'IN_CREATE', 'IN_DELETE', 
    'IN_DELETE_SELF', 'IN_MOVE_SELF', 
    'IN_UNMOUNT', 'IN_Q_OVERFLOW', 
    'IN_IGNORED', 'IN_ONLYDIR', 
    'IN_DONT_FOLLOW', 'IN_EXCL_UNLINK', 
    'IN_MASK_ADD', 'IN_ISDIR', 
    'IN_ONESHOT', 
]

IN_CLOSE=(IN_CLOSE_WRITE | IN_CLOSE_NOWRITE)
IN_MOVE=(IN_MOVED_FROM | IN_MOVED_TO)
IN_ALL_EVENTS=(
    IN_ACCESS | IN_MODIFY | IN_ATTRIB | IN_CLOSE_WRITE |
    IN_CLOSE_NOWRITE | IN_OPEN | IN_MOVED_FROM |
    IN_MOVED_TO | IN_DELETE | IN_CREATE | IN_DELETE_SELF |
    IN_MOVE_SELF
)

IN_CLOEXEC = O_CLOEXEC = 0o02000000
IN_NONBLOCK = O_NONBLOCK = 0o00004000
NAME_MAX = 255 # from limits.h

_libc = ctypes.cdll.LoadLibrary('libc.so.6')
_libc.inotify_add_watch.argtypes = (ctypes.c_int, ctypes.c_char_p, ctypes.c_uint32)
_libc.inotify_add_watch.returntype = ctypes.c_int
_libc.inotify_rm_watch.argtypes = (ctypes.c_int, ctypes.c_int)
_libc.inotify_rm_watch.returntype = ctypes.c_int

class _InotifyHeader(ctypes.Structure):
    """Structure used to unpack the incoming events"""
    _fields_ = [
        ('wd', ctypes.c_int), 
        ('mask', ctypes.c_uint32), 
        ('cookie', ctypes.c_uint32), 
        ('len', ctypes.c_uint32), 
    ]
HEADER_SIZE = ctypes.sizeof(_InotifyHeader)

def raise_on_minus_one( function ):
    """Decorates function to raise InotifyError if result of function is -1
    
    TODO: so far I've never seen this return anything but 0
    from the get_errno() call, it seems likely that the errno
    is getting reset before we get around to it.
    """
    @wraps(function)
    def wrapped(*args, **named):
        result = function(*args, **named)
        err = ctypes.get_errno()
        if result == -1:
            raise InotifyError( err, errno.errorcode.get(err, 'Unknown error'),*args )
        return result 
    return wrapped

class InotifyEvent(object):
    """Encapsulates an Inotify Event instance
    
    Properties:
    
        int:fd -- inotify fd (from init())
        int:wd -- the watch which produced the event
        int:mask -- the event-mask (type) of the event, see the various IN_* constants
                    in the module for the bit-values this can take
        int:cookie -- ID which binds together events that relate to the same underlying file operation
        str:name -- filename to which the event relates
    """
    def __init__(self,fd, wd, mask, cookie, name):
        self.fd = fd 
        self.wd = wd
        self.mask = mask 
        self.cookie = cookie 
        if isinstance(name, bytes):
            name = name.decode(sys.getfilesystemencoding())
        self.name = name
    def get_event_mask(self):
        """Retrieve our event mask as a textual description"""
        keys = globals()
        result = []
        bits = self.mask
        for name in EVENT_NAMES:
            value = keys[name]
            if bits & value:
                result.append(name)
                bits = bits & (~ value)
        if bits:
            result.append(str(bits))
        return "|".join(result)
    @property
    def full_name(self):
        if hasattr(self.wd,'path'):
            if self.name:
                return os.path.join(self.wd.path,self.name)
            else:
                return self.wd.path
        return self.name
    def __repr__(self):
        return '%s(%s,%s,%s,%s,%s)'%(
            self.__class__.__name__, 
            self.fd, 
            self.wd, 
            self.get_event_mask(), 
            self.cookie, 
            clean_repr(self.full_name), 
        )
    @property
    def existing(self):
        """Does this represent an existing file, or a removed one?
        
        There are only a few events that give us "old/stale" names:
        
            IN_DELETE
            IN_DELETE_SELF
            IN_MOVED_FROM
        
        # Special case, IN_MOVE_SELF, IN_DELETE_SELF should update 
        the Inotify OO layer...
        """
        for flag in [IN_DELETE,IN_DELETE_SELF,IN_MOVED_FROM]:
            if self.mask & flag:
                return False 
        return True
    @property
    def isdir(self):
        return self.mask & IN_ISDIR
    @property
    def removed(self):
        return self.mask & IN_MOVED_FROM or self.mask & IN_DELETE

@raise_on_minus_one
def init(flags=O_NONBLOCK):
    """Initialize (produce a file descriptor for watching inotify events)"""
    return _libc.inotify_init1(flags)
@raise_on_minus_one
def add_watch(fd, path, mask=IN_ALL_EVENTS):
    """Add a watch for the given path to the fd"""
    return _libc.inotify_add_watch(fd, as_8_bit(path), mask)
@raise_on_minus_one
def rm_watch(fd, wd):
    """Remove the watch wd from the inotify file descriptor fd
    
    If successful an IN_IGNORED event will be generated on the fd
    """
    result = _libc.inotify_rm_watch(fd, wd)
    if result:
        return result
def get_events(fd, block=None, period=None):
    """Retrieve all events currently available
    
    block -- if None, do not block, only return already-available events
             otherwise pass as the period argument to select.select
             to wait for up to that long for an event to become available
    period -- if None, use default blocking period, otherwise 
    
    yields InotifyEvent instances for all available events
    """
    if block or period:
        return get_events_blocking( fd, poll_period=period or 5.0 )
    else:
        return get_events_nonblocking( fd )
def get_events_blocking(fd, poll_period=5.0):
    """Get pending events, waiting up to poll_period for events to show up"""
    import select 
    if hasattr( select, 'epoll' ):
        poll = select.epoll()
    else:
        poll = select.poll()
    poll.register( fd, select.EPOLLIN|select.EPOLLPRI|select.EPOLLERR|select.EPOLLHUP )
    try:
        while True:
            try:
                for _,_ in poll.poll( poll_period ):
                    for event in get_events_nonblocking( fd ):
                        yield event
                # Orderly exit
                break
            except (IOError,OSError) as err:
                if err.errno not in (errno.EWOULDBLOCK,errno.EINTR):
                    raise
    finally:
        poll.unregister( fd )
    
def get_events_nonblocking(fd):
    """Non-blocking version of get-events, but does *not* wait for events"""
    try:
        read_length = HEADER_SIZE + NAME_MAX + 1
        while True:
            data = os.read(fd, read_length)
            while data:
                header = _InotifyHeader.from_buffer_copy(data)
                name = data[HEADER_SIZE:HEADER_SIZE+header.len]
                if name:
                    name = name[:name.index(NULL_BYTE)]
                data = data[HEADER_SIZE+header.len:]
                yield InotifyEvent(fd, header.wd,header.mask, header.cookie, name) 
    except (IOError,OSError) as err:
        if err.errno not in (errno.EWOULDBLOCK,errno.EINTR):
            raise
    

class Inotify(int):
    """OO facade over the Inotify API
    
    .. doctest::
    
        >>> import os,shutil,tempfile
        >>> temp = tempfile.mkdtemp(prefix='inotify-',suffix='-test')
        >>> fd = Inotify()
        >>> final_dir = os.path.join(temp,'path','to','directory')
        >>> os.makedirs(final_dir)
        >>> wd = fd.add_watch(final_dir,IN_ALL_EVENTS)
        >>> wd.__class__.__name__
        'Watch'
        >>> list(fd.get_events())
        []
        >>> fh = open(os.path.join(final_dir,'test.txt'),'w')
        >>> x = fh.write('oo')
        >>> fh.close()
        >>> for event in fd.get_events():
        ...    print(event) # doctest: +ELLIPSIS
        InotifyEvent(...,...,IN_CREATE,0,'.../inotify-...-test/path/to/directory/test.txt')
        InotifyEvent(...,...,IN_OPEN,0,'.../inotify-...-test/path/to/directory/test.txt')
        InotifyEvent(...,...,IN_MODIFY,0,'.../inotify-...-test/path/to/directory/test.txt')
        InotifyEvent(...,...,IN_CLOSE_WRITE,0,'.../inotify-...-test/path/to/directory/test.txt')
        >>> list(fd.get_events(.001)) # block for up to 1ms
        []
        >>> os.rename(final_dir,final_dir+'.bak')
        >>> for event in fd.get_events():
        ...    print(event) # doctest: +ELLIPSIS
        InotifyEvent(...,...,IN_MOVE_SELF,0,'.../inotify-...-test/path/to/directory')
        >>> fd.rm_watch(wd) # doctest: +SKIP
        1
        >>> fd.add_watch(os.path.join(temp,'moo','roo')) # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
          File "/usr/lib/python2.7/doctest.py", line 1315, in __run
            compileflags, 1) in test.globs
          File "<doctest fussy.inotify[13]>", line 1, in <module>
            fd.add_watch(os.path.join(temp,'moo','roo')) # doctest: +IGNORE_EXCEPTION_DETAIL
          File "/mnt/homevar/var/pylive/fussy/fussy/inotify.py", line 115, in wrapped
            raise InotifyError( err, errno.errorcode.get(err, 'Unknown error') )
        InotifyError: [Errno 0] Unknown error
        >>> fd.close()
        >>> shutil.rmtree(temp)
    
    """
    def __new__(cls,flags=O_NONBLOCK):
        """Create a new Inotify by calling init() and wrapping the result"""
        base = init(flags)
        return super(cls,Inotify).__new__(cls,base)
    def __init__(self):
        """Initialize our internal structures"""
        self.watches = {}
    def add_watch(self,path, mask=IN_ALL_EVENTS):
        """Add a watch for the given path to the Inotify
        
        path -- file-path to add to tracking
        mask -- bitwise or-ing of the flags in which we are interested
        
        returns the wrapped Watch instance for the new watch
        """
        base = add_watch(self,path,mask)
        watch = self.watches[base] = Watch(base,fd=self,path=path,mask=mask)
        return watch
    def rm_watch(self,wd):
        """Remove the watch wd from the inotify file descriptor fd
        
        If successful an IN_IGNORED event will be generated on the fd
        
        wd -- either a simple integer or a Watch instance
        
        returns the internally tracked watch (or wd if not tracked internally)
        """
        try:
            watch = self.watches.pop(wd)
        except KeyError:
            watch = wd
        rm_watch(self,wd)
        return watch
    def get_events(self,*args,**named):
        """Retrieve all events currently available
        
        block -- if None, do not block, only return already-available events
                 otherwise pass as the period argument to select.select
                 to wait for up to that long for an event to become available
        period -- if passed, the duration to wait for an event (assumes blocking)
        
        yields InotifyEvent instances for all available events
        with the fd and wd fields referencing this Inotify and 
        any wd registered with add_watch on this instance.
        """
        for event in get_events(self,*args,**named):
            event.fd = self 
            event.wd = self.watches.get(event.wd,event.wd)
            # TODO: if event.mask & IN_ONESHOT we should de-register
            # the wd after the first cookie we see that uses that 
            # wd is exausted, that is, after all events with that cookie
            # have been seen...
            yield event
    def close(self):
        """Close our file descriptor"""
        os.close(self)

class Watch(int):
    """Wraps watch instances
    
    Simply provides metadata about the watch so that users can see
    the flags, path, etc used in the watch
    """
    def __new__(cls,base,fd,path,mask):
        """Create the new watch instance
        
        base -- base wd (watch id) value
        fd -- the Inotify instance
        path -- our path as passed to add_watch
        mask -- the mask used to create our watch
        """
        base = super(Watch,cls).__new__(cls,base)
        base.fd = fd
        base.path = path 
        base.mask = mask
        return base



"""SSE Client implementation

Needs:

* Reconnecting Client 
* Callback to run on each event
* Default callback should be to mirror the events locally

  * write each (publishable) event to a target/mirror directory (with twrite)

Based on:

    https://github.com/juggernaut/twisted-sse-demo
"""
from twisted.protocols.basic import LineReceiver
from twisted.internet import reactor
from twisted.internet import protocol
from twisted.internet import ssl
import logging,json
try:
    import urlparse 
except ImportError:
    from urllib import parse as urlparse
log = logging.getLogger(__name__)
DEFAULT_URL = 'http://localhost:12000/'

class SSEClientProtocol(LineReceiver):
    def __init__(self):
        self.event = 'message'
        self.data = ''
        self.headers = True
        self.status_code = None
    def connectionMade(self):
        self.transport.write( self.request_headers() )
        
    def request_headers(self):
        return 'GET %s HTTP/1.1\r\n%s\r\n\r\n'%(
            self.factory.final_url(),
            '\r\n'.join([
                '%s: %s'%(k,v) for k,v in [
                    ('Connection','keep-alive'),
                    ('Accept','text/event-stream'),
                    #Last-Event-ID
                ] + self.factory.headers.items()
            ]),
        )

    def lineReceived(self, line):
        """Process a single incoming line"""
        if line.strip() == '':
            if self.headers:
                self.headers = False
            else:
                self.dispatchEvent()
        if self.headers:
            if not self.status_code and line.startswith('HTTP/'):
                self.status_code = int(line.strip().split()[1])
                if self.status_code in (302,401):
                    self.factory.login_first()
            log.info('H: %s',line)
        else:
            #log.info("P: %s",line)
            try:
                field, value = line.split(':', 1)
                # If value starts with a space, strip it.
                value = lstrip(value)
            except ValueError:
                # We got a line with no colon, treat it as a field(ignore)
                return

            if field == '':
                # This is a comment; ignore
                pass
            elif field == 'data':
                self.data += value + '\n'
            elif field == 'event':
                self.event = value.strip()
            elif field == 'id':
                # Not implemented
                self.id = value.strip()
            elif field == 'retry':
                # Not implemented
                # should modify our transport's connection parameters...
                self.transport.initialDelay = float(value.strip())

    def dispatchEvent(self):
        """Dispatch the event"""
        # If last character is LF, strip it.
        if self.data.endswith('\n'):
            self.data = self.data[:-1]
        self.factory.dispatchEvent( self.event, self.data )
        self.data = ''
        self.event = 'message'

class SSEClientFactory(protocol.ReconnectingClientFactory):
    """Reconnecting client that generates SSEClient instances"""
    factor = 1.1
    initialDelay = 1.0
    maxDelay = 20.0
    protocol = SSEClientProtocol
    def __init__(self,*args,**named):
        if 'url' in named:
            self.url = named.pop('url')
        else:
            self.url = DEFAULT_URL
        if 'callbacks' in named:
            self.callbacks = named.pop('callbacks')
        else:
            self.callbacks = {}
        if 'interests' in named:
            self.interests = named.pop('interests')
        else:
            self.interests = '*'
        self.headers = {
            'keep-alive': 3600,
        }
        self.headers.update( named )
    def buildProtocol(self, addr):
        self.resetDelay()
        protocol = SSEClientProtocol()
        protocol.factory = self 
        return protocol
    
    def dispatchEvent(self, event='message', data=''):
        """Dispatch event to one of our handlers if the event is registered"""
        if event in self.callbacks:
            callback = self.callbacks.get(event)
        elif None in self.callbacks:
            callback = self.callbacks.get(None)
        else:
            return 
        try:
            callback( event, data )
        except Exception:
            log.exception("Failure processing %r:%r with function %s",event,data,callback)
    def setFinishedDeferred(self, d):
        self.finished = d
    def addCallback(self, event, func):
        self.callbacks[event] = func
    
    def final_url(self):
        base = self.url 
        if not base.endswith('/'):
            base += '/'
        return base + self.interests

def lstrip(value):
    return value[1:] if value.startswith(' ') else value

def debug_callback(event,data):
    try:
        record = json.loads(data)
    except Exception:
        #log.exception("Failure parsing %r data %r",event,data)
        log.info("%s: %r",event,data)
    else:
        log.info("%s: %s",event,record)

class NoverifyFactory(ssl.ClientContextFactory):
    def getContext(self):
        from OpenSSL import SSL
        self.method = SSL.SSLv23_METHOD
        ctx = ssl.ClientContextFactory.getContext(self)
        ctx.verify = False
        return ctx


def client( url, callbacks=None ):
    if not callbacks:
        callbacks = { 
            None:debug_callback,
        }
    parsed = urlparse.urlparse(url)
    factory = SSEClientFactory( callbacks=callbacks, url=parsed.path, host=parsed.hostname )
    if parsed.scheme == 'httpsu':
        reactor.connectSSL(parsed.hostname,parsed.port or 443,factory,NoverifyFactory())
    elif parsed.scheme == 'https':
        reactor.connectSSL(parsed.hostname,parsed.port or 443,factory)
    elif parsed.scheme == 'http':
        reactor.connectTCP(parsed.hostname,parsed.port or 80,factory)
    return factory

def get_options():
    import argparse
    parser = argparse.ArgumentParser(description='Run SSE client demo (Twisted Inotify Publishing Channels)')
    parser.add_argument(
        '-u','--url', metavar='URL',
        default = DEFAULT_URL,
        help="http url to which to connect to the server, default: %s"%(DEFAULT_URL,),
    )
    
    return parser

def main():
    logging.basicConfig(level=logging.INFO)
    options = get_options().parse_args()
    factory = client( options.url )
    assert factory
    reactor.run()

if __name__ == "__main__":
    main()



#! /usr/bin/env python
import json,sys
def main():
    if sys.argv[1:]:
        source = open(sys.argv[1]).read()
    else:
        source = sys.stdin.read()
    content = json.loads(source)
    sys.stdout.write(
        json.dumps( 
            content, 
            indent=2,
            sort_keys=True,
        )
    )
if __name__ == "__main__":
    main()



"""Server Sent Events Channel for Fussy

This is a push service that is intended for use in very small
servers (such as fussy is normally used to manage). It does 
*not* attempt to scale up to huge installations.

Based on this code:

    https://github.com/juggernaut/twisted-sse-demo

Example usage (Note: you may need a shim to get this to work in all browsers,
such as: https://github.com/Yaffle/EventSource ):

    var eventSource = new EventSource("http://localhost:12000/test2.txt");
    eventSource.addEventListener('test2.txt',function(event) {
        element = $(".moo-test");
        element.text(event.type+' '+event.lastEventId+' '+event.data);
    });

Now to send an event:

    echo "Moo" > /var/firmware/protected/test2.txt

and the event will be formatted into the web-page

Nginx stanza to include the server:

    location /__sse__/ {
        # Protected by the web app, but served by the server...
        internal;
        proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
        proxy_set_header X-Nginx-Hosted "Yes";
        proxy_set_header Host $http_host;
        proxy_set_header X-Forwarded-Protocol $scheme;
        proxy_read_timeout 1200s;
        proxy_send_timeout 1200s;
        proxy_redirect off;
        
        proxy_pass http://localhost:12000/;
        break;
    }
    
Then in Django (web/app) you do an NGINX Redirect *after*
you have decided that a given user is allowed to view *all* 
channels in the sse_server:

    @permission_required('yada.yoo')
    def foo( request, interest ):
        # validate rights to get to the these channels
        if not channel_permissions( interest ):
            return HttpResponse( 'Permissions prevent listening to those channels', 401)
        if request.META.get( 'HTTP_X_NGINX_HOSTED' ):
            response = HttpResponse( '' )
            response['X-Accel-Redirect'] = '/__sse__/%s'%(urllib.quote_plus(interest))
            return response
        else:
            return HttpResponseRedirect( '%s/%s'%(settings.SSE_SERVER_DEBUG_LOCATION,%(urllib.quote_plus(interest) )

Clients can then connect and process the SSEvents at whatever URL 
you have put the bit of view code there.
"""
import re, sys, urllib, weakref
from twisted.web import server, resource
from twisted.internet import reactor
from twisted.python import log
from twisted.internet import inotify
from twisted.python import filepath
from twisted.internet import task
from twisted.internet import endpoints
from fussy import tailprotocol

PUBLISHABLE = re.compile(r'^([a-zA-Z0-9][a-zA-Z0-9-.*]{0,64})|[*]$')
DEFAULT_PUBLISH_ROOT = '/var/firmware/protected/live'
DEFAULT_ENDPOINT = 'tcp:12000:interface=127.0.0.1'
DEFAULT_KEEPALIVE = 20.0

class RootMonitor( object ):
    def __init__(self,root, server):
        self.root = root 
        self.server = server
        self.server.monitor = self
        self.start_watching()
    def start_watching(self):
        notifier = inotify.INotify()
        notifier.startReading()
        notifier.watch(
            filepath.FilePath(self.root), 
            mask = inotify.IN_CLOSE_WRITE|inotify.IN_MOVED_TO|inotify.IN_MODIFY,
            callbacks=[self.on_file_change]
        )
    @classmethod
    def publishable(cls, name ):
        return bool(PUBLISHABLE.match(name))
    
    def on_file_change(self,ignored, filepath, mask):
        """A file has changed, see if it's a publishable event"""
        base = filepath.basename()
        log.msg('Update to %s'%(base,))
        if self.publishable( base ):
            if mask & inotify.IN_MODIFY:
                # tail operations only...
                self.server.update_tails( filepath )
            else:
#                if mask & inotify.IN_CLOSE_WRITE:
#                    self.server.close_tails( filepath )
                try:
                    content = filepath.open().read().splitlines()
                except Exception:
                    log.msg("Unable to read %s"%(filepath,))
                else:
                    self.server.send_all( base, content )
    def current(self):
        """Return the current paths/events known to the server"""
        root = filepath.FilePath(self.root)
        return sorted(root.globChildren('*'), key=lambda x: x.getModificationTime())

class SSEServer(resource.Resource):
    """SSE Server that publishes files from a directory"""
    isLeaf = True
    def __init__(self, keepalive=DEFAULT_KEEPALIVE):
        self.subscribers = {}
        self.tails = {}
        self.event_id = 1
        self.keepalive = keepalive
        task.LoopingCall( self.on_keepalive ).start( self.keepalive )
    _monitor = None
    @property
    def monitor(self):
        return self._monitor()
    @monitor.setter
    def monitor(self, other):
        self._monitor = weakref.ref(other)
    
    def on_keepalive(self):
        """Send a keepalive (ping) message"""
        for subscriber,interests in self.subscribers.items():
            try:
                subscriber.write(":\n")
            except Exception:
                pass
        
    def send_all(self, event, content, subscribers=None ):
        log.msg("Sending event: %r"%(event,))
        if not self.subscribers:
            # don't waste ids...
            log.msg("No current subscribers")
            return
        subscribers = [
            (s,i) for s,i in (subscribers or self.subscribers).items()
            if self.interested( event, i )
        ]
        if not subscribers:
            log.msg("No one interested in %s"%(event,))
            return
        self.event_id += 1
        self.event_id %= 0xffffff
        for subscriber,interests in subscribers:
            log.msg('Sending %s lines to %s'%(len(content),subscriber,))
            try:
                if interests:
                    subscriber.write("event: %s\n"%(event,))
                subscriber.write("id: %s\n"%(self.event_id,))
                for line in content:
                    subscriber.write("data: %s\n" % line)
                # NOTE: the last CRLF is required to dispatch the event at the client
                subscriber.write("\n")
            except Exception as err:
                log.msg( "Failed sending to %s: %s"%( subscriber, err) )

    def interested(self, base, interests ):
        """Is this set of interests interested in base event?"""
        if (not interests) or '*' in interests:
            # interested in everything...
            return True 
        else:
            return base in interests
    def update_tails(self, filepath ):
        """Update any items which are tailing filepath.basename"""
        base = filepath.basename()
        for subscriber in self.subscribers.keys():
            map = getattr(subscriber,'tails',{})
            log.msg("tails: %s"%(map))
            if base in map:
                tail = map.get(filepath)
                if tail is None:
                    log.msg( 'Starting new tail for %s'%(filepath.path))
                    tail = map[filepath] = tailprotocol.FileTail( filepath )
                lines = tail.read_from_file()
                if lines:
                    self.send_all( base, lines, {subscriber:{base:True}} )
    def render_GET(self, request):
        for header,value in [
            ('Content-Type', 'text/event-stream; charset=utf-8'),
            ('Access-Control-Allow-Origin','*'),
            ('Connection','keep-alive'),
            ("Cache-Control", "no-cache"),
        ]:
            request.setHeader(header,value)
        request.setResponseCode(200)
        log.msg( "Connection from %s"%(request.getClientIP()))
        d = request.notifyFinish()
        d.addBoth(self.on_disconnect)
        self.subscribers[request] = interests = set()
        for fragment in request.postpath:
            fragment = urllib.unquote_plus(fragment)
            for interest in fragment.split(' '):
                if interest.startswith('-'):
                    interest = interest[1:]
                    if RootMonitor.publishable(interest):
                        if not getattr(request,'tails',None):
                            request.tails = {}
                        # set of tails for the interest (indexed by path)
                        request.tails.setdefault( interest, {} )
                if RootMonitor.publishable( interest ):
                    interests.add(interest)
        request.write("")
        reactor.callLater(0, self.send_initial, request )
        return server.NOT_DONE_YET
    def send_initial(self, subscriber ):
        interests = self.subscribers[subscriber]
        # do *initial* sends for each of the interests...
        for existing in self.monitor.current():
            base = existing.basename()
            if self.interested( base, interests ):
                try:
                    content = existing.open().read().splitlines()
                except Exception:
                    log.msg("Unable to read %s"%(existing,))
                else:
                    self.send_all( base, content, {subscriber:interests} )
#            elif base in self.logs and subscriber in self.logs[base]:
#                size = existing.getsize()
#                content = existing.open().read( size )
#                lines = content.splitlines()
#                if lines and lines[-1] == '':
#                    lines = lines[:1]
#                else:
#                    size = size-len(lines[-1])
#                    lines = lines[:-1]
#                self.send_all( base, lines, {subscriber:

    def on_disconnect(self, subscriber):
        log.msg('Disconnect: %s'%(subscriber))
        if subscriber in self.subscribers:
            if not subscriber.finished:
                subscriber.finish()
            log.msg("Removing subscriber..")
            try:
                del self.subscribers[ subscriber ]  
            except KeyError:
                log.info("Unable to remove %s"%(subscriber,))
            for name, interested in self.logs.items():
                if subscriber in interested:
                    del interested[subscriber]
                if not interested:
                    del self.logs[name]
    
def get_options():
    import argparse
    parser = argparse.ArgumentParser(description='Run SSE server (Twisted Inotify Publishing Channels)')
    parser.add_argument(
        '-r','--root', metavar='DIR',
        default = DEFAULT_PUBLISH_ROOT,
        help="Directory to be published, default: %s"%(DEFAULT_PUBLISH_ROOT,),
    )
    parser.add_argument(
        '-e','--endpoint', metavar='ENDPOINT',
        default = DEFAULT_ENDPOINT,
        help="Twisted-style endpoint on which to publish, default: %s\nSee: http://twistedmatrix.com/documents/current/api/twisted.internet.endpoints.serverFromString.html"%(DEFAULT_ENDPOINT,),
    )
    parser.add_argument(
        '-k','--keepalive',metavar='SECONDS',
        default=DEFAULT_KEEPALIVE,
        help="Period at which keepalive messages will be sent",
    )
    parser.add_argument(
        '-l','--logfile',metavar='FILE',
        default = None,
        help="Log file to write, default: stdout",
    )
    return parser

def main():
    options = get_options().parse_args()
    if options.logfile:
        log_target = open(options.logfile,'w')
    else:
        log_target = sys.stdout
    log.startLogging(log_target)
    log.msg( "Publishing %s on %s"%(
        options.root,
        options.endpoint,
    ))
    sse = SSEServer( keepalive = options.keepalive )
    root = RootMonitor( options.root, sse )
    assert root
    site = server.Site(sse)
    endpoint = endpoints.serverFromString(reactor,options.endpoint )
    endpoint.listen(site)
    reactor.run()

if __name__ == "__main__":
    main()



"""Link all files in a distribution from relative location on the hard disk"""
import os, logging, shutil
from optparse import OptionParser

log = logging.getLogger( __name__ )

def link_tree( distribution_dir, target_dir='/', symbolic=True ):
    """(Sym)link all files under distribution_dir from target_dir 
    """
    if symbolic:
        link = os.symlink 
    else:
        link = os.link
    for path, subdirs, filenames in os.walk( distribution_dir ):
        dir_relative = os.path.relpath( path, distribution_dir )
        dir_source = os.path.normpath( os.path.abspath( path ) )
        dir_final = os.path.normpath( os.path.join( target_dir, dir_relative ))
        if not os.path.exists( dir_final ):
            log.info( 'Creating directory: %s', dir_final )
            try:
                os.makedirs( dir_final )
            except Exception as err:
                err.args += (dir_final,)
                raise 
        for filename in filenames + subdirs:
            file_source = os.path.join( dir_source, filename )
            file_final = os.path.join( dir_final, filename )
            if os.path.islink( file_source ):
                
                target = os.readlink( file_source )
                final_target = os.path.join( dir_final,target)
                log.info('Recreating link in %s to %s',file_final, final_target )
                create_link( final_target, file_final, os.symlink )
            elif os.path.isfile(file_source):
                if os.path.lexists( file_final ):
                    if (
                        symbolic and 
                        os.path.islink( file_final ) and 
                        os.readlink( file_final ) == file_source
                    ):
                        log.info( 'Up to date: %s', file_final )
                        continue
                    log.info( 'Replacing: %s', file_final )
                    create_link( file_source, file_final, link )
                else:
                    log.info( 'Linking: %s', file_final )
                    create_link( file_source, file_final, link )

def create_link( file_source, file_final, link ):
    try:
        temp = file_final + '~'
        if os.path.exists( temp ) or os.path.lexists( temp ):
            try:
                os.remove( temp )
            except Exception as err:
                err.args += (temp, )
                raise 
        try:
            link( file_source, temp )
        except OSError as err:
            if link is os.link:
                shutil.copy2( file_source, temp )
            else:
                raise
        os.rename( temp, file_final )
    except Exception as err:
        log.error( "Error linking %s: %s", file_final, err )
        raise err
    
def get_options():
    """Creates the OptionParser used in :func:`main` """
    parser = OptionParser()
    parser.add_option(
        '-d','--directory',
        dest = 'directory',
        default = None,
        help="Directory to which to link",
    )
    parser.add_option(
        '-t','--target',
        dest = 'target',
        default = None,
        help="Directory in which to create the links",
    )
    parser.add_option(
        '-v','--verbose',
        dest = 'verbose',
        action="store_true",
        default = False,
        help="Perform verbose logging (INFO level), default: False",
    )
    parser.add_option(
        '-H','--hard',
        dest = 'hardlink',
        action="store_true",
        default = False,
        help="Use hard-links instead of symlinks, default: False",
    )
    return parser

def main():
    options,args = get_options().parse_args()
    if options.verbose:
        logging.basicConfig( level=logging.INFO )
    else:
        logging.basicConfig( level=logging.WARNING )
    if args:
        if not options.directory:
            options.directory = args[0]
            args = args[1:]
        if args and not options.target:
            options.target = args[0]
            args = args[1:]
    link_tree( options.directory, options.target, symbolic = (not options.hardlink) )
    return 0



"""Utility functions for common tasks"""
from fussy import install, nbio
import os, glob, logging
from optparse import OptionParser
log = logging.getLogger( __name__ )

MIGRATION_PATTERN = '????_*.py'

def find_reverse_upgrades( final_target, relative_path, migration_pattern=MIGRATION_PATTERN ):
    """Find reverse upgrades that must be run before final_target is installed
    
    final_target -- final target of the installation
    relative_path -- path from final target to Django/South migrations
    
    returns ID/number of the migration to run (if any)
    """
    new = glob.glob( os.path.join( final_target, relative_path, migration_pattern ) )
    old = glob.glob( os.path.join( final_target, '..', install.CURRENT_LINK, relative_path, migration_pattern ) )
    new = sorted([os.path.basename( x ) for x in new ])
    old = sorted([os.path.basename( x ) for x in old ])
    # we want the last migration in both old and new
    common = [ old_item for (new_item,old_item) in zip( new,old ) if new_item == old_item ]
    if common == old:
        # forward migration only 
        log.info( 'No reverse migration required' )
        return None
    elif common:
        # there is a common set, but it is *not* the same as old, so roll back current 
        # until it is at the common branching point...
        log.warn(
            "Extra migrations in current versus target: %s",
            ", ".join( old[len(common):] ),
        )
        return os.path.splitext( common[-1] )[0]
    else:
        # I can't think of a legitimate use case for wanting to roll back to having no data...
        raise RuntimeError( "No common migrations found" )
def get_options():
    """Creates the OptionParser used in :func:`main` """
    parser = OptionParser()
    parser.add_option(
        '-f', '--final_target',
        dest = 'final_target',
        help = "Path to the new firmware installation",
    )
    parser.add_option(
        '-d','--django-admin',
        dest = 'django_admin',
        default = 'django-admin.py',
        help="Path to the django-admin.py script with which to run migrations",
    )
    parser.add_option(
        '-p', '--path',
        dest = 'path',
        help = "Relative path from current to migration directory...",
    )
    parser.add_option(
        '-a', '--app',
        dest = 'app',
        help = "Application name to migrate",
    )
    return parser

def main():
    """Finds last common migration and reverts current Django db to that 
    
    Must be run with DJANGO_SETTINGS_MODULE set in the environment
    """
    logging.basicConfig( level=logging.INFO )
    options,args = get_options().parse_args()
    migration = find_reverse_upgrades( 
        options.final_target,
        options.path,
    )
    if migration:
        pipe = nbio.Process([
            options.django_admin,
            'migrate',
            '--noinput',
            options.app,
            migration,
        ],stderr=-1)
        print(pipe())



#! /usr/bin/env python
"""Unpack a signed .tar.gz.asc or .tar.gz.gpg archive into a temporary directory
"""
import datetime

def ts_name( ):
    """Produce a human-readable timestamp (UTC ISO formatted date)"""
    d = datetime.datetime.utcnow()
    return d.isoformat().replace(':','.')



import os,threading,time,logging,signal,errno
from fussy import nbio
log = logging.getLogger( __name__ )

__all__ = [
    'alive',
    'descendants',
    'running_pids',
    'unmanaged',
    
    'kill_kill_kill',
    'kill_children',
    'kill_unmanaged',
    'kill_by_name',
    
    'daemon_thread',
    'wait',
    'current_time',
]

def alive( pid ):
    """Is the given PID responding to signals?"""
    try:
        os.kill( pid, 0 )
    except OverflowError:
        raise ValueError( pid, "Is too large for a pid value")
    except OSError as err:
        if err.args[0] == errno.ESRCH: # no such process
            return False
        elif err.args[0] == errno.EPERM: # likely owned by someone else
            log.warn( "Attempted to send signal to process we cannot signal" )
            return False 
        log.debug( 'Unexpected return code: %s', err )
        return False 
    else:
        return True

def _pgrep_pids( pids ):
    """Cleans up PID listing from pgrep output"""
    return [int(x) for x in pids.strip().splitlines()]    

def descendants( pid=None ):
    """Get all descendants of the current PID
    
    Note: this is a *recursive* calling of a pgrep operation, it is 
    *not* intended for use with extremely deep trees
    
    yields children, then parent
    """
    pid = pid or os.getpid()
    parents, children = proc_map()
    return list(_depth_first(children, pid))

def _depth_first(map, start):
    for child in map.get(start, ()):
        yield child 
        for descendant in _depth_first(map, child):
            yield descendant

def proc_map():
    """Pull out a process map of all processes on the system"""
    procs = [
        x for x in os.listdir('/proc') if x.isdigit()
    ]
    children = {}
    parent = {}
    for proc in procs:
        try:
            fh = open(os.path.join('/proc', proc, 'status'))
        except Exception:
            pass 
        else:
            try:
                for line in fh:
                    if line.startswith('PPid:'):
                        proc, ppid = int(proc, 10), int(line.split(':')[1].strip(), 10)
                        parent[proc] = ppid 
                        children.setdefault(ppid, []).append(proc)
            finally:
                fh.close()
    return parent, children


def kill_children( pid=None ):
    """Kill all children of this process
    
    pid -- if not specified, this process (which isn't very useful, as it 
        will kill the code running this operation)
    
    kill_kill_kill's for each descendants( pid ) and then kill_kill_kill( pid )
    """
    for child in list(descendants( pid )):
        try:
            kill_kill_kill( child )
        except (IOError,OSError):
            # race condition where the child has already exited...
            pass 
    kill_kill_kill( pid )

def describe( pid ):
    """Produce a short description of the PID for use in log messages"""
    try:
        return nbio.Process( ['ps','f','--pid', str(pid)])().decode('utf-8').strip().splitlines()[1]
    except (IndexError,nbio.ProcessError):
        return u'Process #%s'%(pid,)
    
def kill_kill_kill( pid ):
    """Kill pid (with prejudice)
    
    Attempts three times to kill the process with three signals (INT, TERM, KILL)
    for a total of 9 kill attempts.  Exits as soon as alive( pid ) is False.
    
    Waits .1* 2**(attempt) between each signal.
    
    Used when you absolutely *must* be sure that those processes are gone.
    """
    if alive( pid ):
        for attempt in range( 3 ):
            for sig in [signal.SIGINT,signal.SIGTERM,signal.SIGKILL]:
                try:
                    os.kill( pid, sig )
                except OSError as err:
                    if err.args[0] == errno.ESRCH:
                        # yay, it's gone!
                        return True 
                    elif err.args[0] == errno.EPERM:
                        # we can't send signals, no point retrying...
                        return False 
                    else:
                        log.info(
                            "Kill returned unexpected failure: %s", err 
                        )
                wait( 0.1* (2**attempt) )
                if not alive( pid ):
                    return True 
                log.debug( 'PID %s still alive after attempted kill with %s', pid, sig )
    return False

def running_pids( name ):
    """Return pids of all instances of processes named (precisely) name
    
    Uses an exact match on the name
    """
    pids = nbio.Process([
        'pgrep', '-x', name,
    ], good_exit = [0,1])()
    return _pgrep_pids( pids )
def kill_by_name( name ):
    """Kill every process in running_pids( name )"""
    for pid in running_pids( name ):
        log.info( "Killing %s instance %s", name, pid )
        kill_kill_kill( pid )
def unmanaged( name ):
    """Return pids of all instances of processes named name which do not have parents
    
    More precisely, those instances where the process is a child of PID 1.  
    This is used to kill off processes that should normally be run under 
    supervisor or another management/scheduling daemon, but which have now 
    "escaped" due to the death of the parent process.
    
    Matching is by partial match against the whole command line, so 
    'thunder' would match 'thunderbird' and python script names are visible
    """
    pids = nbio.Process([
        'pgrep', '-P','1', '-f', name
    ], good_exit=(0,1))()
    return _pgrep_pids( pids )
def kill_unmanaged( name ):
    """Kill all unmanaged instances of the given process name (and their descendants)
    
    kill_children(pid) for pid in unmanaged( name )
    """
    for pid in unmanaged( name ):
        log.info( "Killing unmanaged %s instance and its children: %s", name, pid )
        kill_children( pid )

def daemon_thread( target, **named ):
    """Create a daemon thread from the named parameters"""
    if 'name' not in named:
        # by default, give the threads names that match the target...
        try:
            named['name'] = getattr(target,'__name__')
        except Exception:
            pass
    thread = threading.Thread( target=target, **named )
    thread.daemon = True 
    assert thread.daemon
    log.info( 'Starting background thread for %s', getattr(target,'__name__',target))
    thread.start()
    assert thread is not None
    return thread 

def wait( duration ):
    """Wait a given duration for some event (for testing, allows us to simulate time)"""
    return time.sleep( duration )
    
def current_time():
    """Time retrieval point (for testing, allows us to simulate time)"""
    return time.time()




"""Unpack a signed .tar.gz.asc or .tar.gz.gpg archive into a temporary directory
"""
import tempfile, os, shutil, glob, logging, traceback
from optparse import OptionParser
from fussy import errors, nbio
log = logging.getLogger( __name__ )

DEFAULT_KEYRING = '/etc/fussy/keys'

def verify_requirements( directory, requirements ):
    """Verify that the directory has the expected files
    
    requirements -- list of files expected to exist in the directory,
        should be minimal, we've already crypto-verified the source, 
        this is just to prevent signed-but-defective installations
    
    raises 
    """
    for requirement in requirements:
        if not os.path.exists( os.path.join( directory, requirement )):
            raise RuntimeError( "Missing path: %r"%( requirement, ))

def unpack( filename, keyring=DEFAULT_KEYRING ):
    """Unpack (uploaded) filename into temporary directory
    """
    source = os.path.abspath( os.path.normpath( filename ) )
    directory = tempfile.mkdtemp( prefix='fussy-', suffix='-unpacked' )
    try:
        env = os.environ.copy()
        env['GNUPGHOME'] = keyring
        try:
            temptar = os.path.join( directory, 'firmware.tar.gz' )
            nbio.Process( ['gpg','-o',temptar,'-d', source], env=env, stderr=False )()
            nbio.Process( ['tar','-zxf', temptar], cwd=directory, stderr=False )()
        except nbio.ProcessError as err:
            #log.error( "Traceback: %s", traceback.format_exc())
            try:
                output = '\n'.join(getattr(err.process,'output',[]))
            except Exception as err:
                output = ''
            if err.process.command[0] == 'gpg':
                log.error( 'gpg reported failure: %s', output )
                raise errors.SignatureFailure( 
                    "Unable to verify signature"
                )
            else:
                log.error( 'tar reported failure: %s', output )
                raise errors.ArchiveFailure(
                    "Unable to unpack the archive: %s"%( str(err))
                )
        except Exception as err:
            log.error( "Unhandled error Traceback: %s", traceback.format_exc())
            raise 
        else:
            os.remove( temptar )
        # there should now be *precisely* one firmware directory and 
        # potentially 2 upgrade scripts (which don't match *)
        files = glob.glob( os.path.join( directory, '*' ))
        if len(files) != 1:
            raise errors.ArchiveFailure( "Expected a single root directory, got: %s"%(files,))
        directory = files[0]
        if os.path.basename( directory ) in ['current']:
            raise errors.ArchiveFailure( "Do not allow creation of archives with name current" )
    except Exception as err:
        shutil.rmtree( directory, True )
        raise
    return directory

def get_options():
    parser = OptionParser()
    parser.add_option(
        '-f','--file',
        dest = 'file',
        default = None,
        action="store",
        type="string",
        help="The firmware archive to unpack, must be a .tar.gz.gpg or a .tar.gz.asc",
    )
    parser.add_option(
        '-k','--keyring',
        dest = 'keyring',
        default = '/etc/fussy/keys',
        action="store",
        type="string",
        help="GPG keyring to use for verification/decryption (default /etc/fussy/keys)",
    )
    parser.add_option(
        '-v', '--verbose',
        dest = 'verbose',
        action = 'store_true',
        default = False,
        help = 'Do verbose logging',
    )
    return parser
    
def main():
    parser = get_options()
    options,args = parser.parse_args()
    logging.basicConfig( level=[logging.INFO,logging.DEBUG][bool(options.verbose)] )
    if not options.file:
        if args:
            options.file = args[0]
        else:
            parser.error( "Need a file to unpack" )
    filename = options.file or args[0]
    log.info( 'Unpacking: %s', filename )
    log.info( 'Keyring: %s', options.keyring )
    directory = unpack( filename, keyring=options.keyring )
    print(directory)
    return 0



"""Errors raise by Fussy explicitly
"""
class FussyError( Exception ):
    """Base class on which all other Fussy specific errors are based"""
    EXIT_CODE = 2

class SystemSetupError( RuntimeError, FussyError ):
    """System configuration appears incorrect (missing required packages, etc)
    
    Recovery will likely require system-level intervention
    """
    EXIT_CODE = 40

class RevertedFailure( FussyError ):
    """Error when failure occurred, but were able to revert to previous/failsafe
    
    No recovery is required, the previous version of the firmware was restored
    """
    EXIT_CODE = 3
class UnrecoverableError( FussyError ):
    """Error raised when there's just no way to recover"""
    EXIT_CODE = 4

class SignatureFailure( FussyError ):
    """Signature was incorrect, or could not be verified
    
    No recovery is required, the archive was rejected before install
    """
    EXIT_CODE = 20
class ArchiveFailure( FussyError ):
    """Archive appears incorrectly structured
    
    No recovery is required, the archive was rejected before install
    """
    EXIT_CODE = 21



"""JSON log server (log-line server)

Logs each incoming log line to the disk *without modification*, but 
only writes lines when the line is complete, so allows you to log from 
many processes simultaneously.

See :py:class:`fussy.jsonlog.JSONSocketHandler` for how to send messages 
to the log server.

The log server should be set up to run under a process manager 
(e.g. supervisord) like so::

    [program:fussylog]
    command=/opt/firmware/current/env/bin/fussy-log-server -f /var/www/logs/errors.json
    autorestart=true
    user=<someuser>
    directory=/var/www/logs
    stopasgroup=true
    killasgroup=true 
    startretries=200
    priority=50
"""
import asyncore
import socket
import json
import os 
import sys
import time
import logging,traceback
log = logging.getLogger( __name__ )
from optparse import OptionParser

class LogHandler(asyncore.dispatcher):
    """Handle messages from a single client process"""
    buffer = ''
    def __init__( self, *args, **named ):
        self.sink = named.pop( 'sink' )
        asyncore.dispatcher.__init__( self, *args, **named )
    def handle_read(self):
        try:
            self.buffer += self.recv(8192)
            i = 0
            while '\n' in self.buffer and i < 50:
                index = self.buffer.index( '\n' )
                record,self.buffer = self.buffer[:index+1],self.buffer[index+1:]
                self.sink( record )
                i += 1
            if i >= 50:
                log.warn( "More than 50 records received at the same time!" )
        except Exception:
            log.error( "Unhandled error on read: %s", traceback.format_exc() )
            self.close()
            raise
    def handle_close(self):
        self.close()
    def writable( self ):
        return False

class LogServer(asyncore.dispatcher):
    """Handle incoming connections from however many processes wish to log"""
    def __init__(self, host='127.0.0.1', port=2514, sink=None):
        asyncore.dispatcher.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.sink = sink
        self.bind((host, port))
        self.listen(5)

    def handle_accept(self):
        try:
            pair = self.accept()
            if pair is None:
                pass
            else:
                sock, addr = pair
                log.info( "New connection from: %s", addr )
                LogHandler(sock, sink=self.sink)
        except Exception:
            log.error( "Unhandled error on accept: %s", traceback.format_exc())
            raise

ROTATION = 1024*1024
ROTATION_COUNT = 5
class Sink( object ):
    """Location to which to send messages (rotating files in the simple case)
    """
    file_handle = None
    def __init__( self, filename, max_bytes = ROTATION, max_rotation=ROTATION_COUNT ):
        """Create a sink for the given filename"""
        self.filename = filename 
        self.max_bytes = max_bytes
        self.max_rotation = max_rotation
        if os.path.exists( self.filename ):
            self.written = self.last_flush = os.stat( self.filename ).st_size
            self.file_handle = open( self.filename, 'ab' )
        else:
            self.written = 0
            self.last_flush = 0
            self.file_handle = open( self.filename, 'wb' )
        self.file_handle.write('\n')
        self.written += 1
        #self.rotate()
    def rotate( self ):
        log.info( 'Rotating: %s', self.filename )
        if self.file_handle:
            self.file_handle.close()
        self.written = 0
        self.last_flush = 0
        if os.path.exists( self.filename ):
            for i in range( self.max_rotation-1, 0, -1 ): # stop at 1
                rot = '%s.%s'%( self.filename, i )
                rotpone = '%s.%s'%( self.filename, i+1 )
                if os.path.exists( rot ) and os.stat( rot ).st_size > 0:
                    os.rename( rot, rotpone )
            i = 1
            rot = '%s.%s'%( self.filename, i )
            os.rename( self.filename, rot )
        self.file_handle = open( self.filename, 'wb' )
    # By default, flush on every write, not good on a laptop, but we're logging 
    # high-priority messages with this thing...
    FLUSH_SIZE = 0
    def __call__( self, record ):
        self.written += len(record)
        if self.written > self.max_bytes:
            self.rotate()
            self.written = len(record)
        self.file_handle.write( record )
        if self.written - self.last_flush > self.FLUSH_SIZE:
            self.file_handle.flush()
            self.last_flush = self.written

LOGGING_PORT = 2514
def server_options():
    parser = OptionParser()
    parser.add_option(
        '-f','--file',
        dest = 'filename',
        default = '/tmp/json.log',
        help="File into which to log",
    )
    parser.add_option(
        '-r', '--rotation',
        dest = 'rotation',
        default = ROTATION,
        help = "Size after which to rotate log file (bytes), defaults %s"%( ROTATION,),
    )
    parser.add_option(
        '-c', '--count',
        dest = 'rotation_count',
        default = ROTATION_COUNT,
        help = "Number of rotated logs to retain, default %s"%( ROTATION_COUNT, ),
    )
    parser.add_option(
        '-p', '--port',
        dest = 'port',
        default = 2514,
        type = 'int',
        help = 'Port on which to listen',
    )
    return parser

def main():
    """Sink json-formatted messages into a rotating file handle"""
    parser = server_options()
    options,args = parser.parse_args()
    directory = os.path.dirname( options.filename )
    if not os.path.exists( directory ):
        os.makedirs( directory )
    logging.basicConfig( 
        filename=options.filename + '.logserver.log', level=logging.INFO 
    )
    sink = Sink( 
        filename=options.filename, 
        max_bytes=options.rotation, 
        max_rotation=options.rotation_count,
    )
    log.info( 'Listening on port %s and writing to file %s', options.port, options.filename )
    LogServer( sink = sink, port=options.port )
    try:
        asyncore.loop()
    except KeyboardInterrupt:
        return 0

def format_date( ts ):
    return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(ts))
def format_record( record ):
    record['date'] = format_date(record['created'])
    record['message'] = '   '.join( record['msg'].splitlines())
    return u'%(date)s %(levelname)s %(name)s: (%(filename)s:%(lineno)s %(funcName)s)    %(message)s'%record
def format_log(filename):
    """Format a given fussy log server file into more traditional format"""
    for line in open( filename ):
        try:
            if line.strip():
                print(( format_record( json.loads(line.strip())).encode('utf-8')))
        except Exception:
            log.warn('Failed to process line: %r',  line)
def format_log_main():
    if sys.argv[1:]:
        for filename in sys.argv[1:]:
            format_log( filename )
        return 0
    else:
        print( 'fussy-format-log filename [filename...]' )
        return 1
    
if __name__ == "__main__":
    main()



"""JSONLogServer client for Python logging

To install a JSON Log Server client (logger), do the following::

    from fussy.jsonlog import JSONSocketHandler
    
    root = logging.getLogger( '' )
    handler = JSONSocketHandler( )
    # important, the logger is *not* intended for high volumes of 
    # messages, it is intended for messages you might actually want to 
    # look at, so set the level high!
    handler.setLevel( logging.WARN )
    root.addHandler( handler )

To add to a Django logging config::

    'handlers': {
        'json': {
            'level': 'WARN',
            'class': 'fussy.jsonlog.JSONSocketHandler',
        },
    },
    'loggers': {
        '': {
            'handlers': ['json'],
        }
    }

Records stored look like this:

    {
      "name": "root", # logger name
      "created": 1354250362.095767, # unix timestamp (time.time())
      "args": [], # unicode(a) for a in args
      "module": "<stdin>", # python module from which the message comes
      "funcName": "<module>", # function from which the message was generated
      "levelno": 30, # numeric code for the level
      "pathname": "<stdin>", # path to the file/module
      "lineno": 1, # line at which the logging message is produced
      "msg": "Hello world", # message, with *no* substitutions performed!
      "filename": "<stdin>", # the file name for the code generating the error
      "levelname": "WARNING" # the textual name of the level
    }
    
    
"""
import logging 
import logging.handlers
import socket
import json
import time
import traceback
import threading
LOGGING_PORT = 2514

class JSONSocketHandler( logging.handlers.SocketHandler ):
    """Handler which sends JSON-formatted subsets of logging records to a socket
    
    The handler actually uses a JSONFormatter to format the records, 
    and you can override the formatter by doing a ``setFormatter`` on the 
    handler, should you wish to override the format logged.
    
    Note that messages *must* be single-line, as the log server uses newlines 
    as the indication that a record may be written (so if you aren't single line 
    your messages can become scrambled).
    """
    def __init__( self, host='127.0.0.1',port=LOGGING_PORT ):
        logging.handlers.SocketHandler.__init__( self, host, port )
        if not self.formatter:
            self.formatter = JSONFormatter()
        # we really, really, really want to get messages through, and 
        # the log server should *always* be up, so be far more aggressive 
        # about retrying to send messages...
        self.retryStart = .01
        self.retryMax = 5.0
        self.retryFactor = 2.0
        self.buffer = []
        self.lock = threading.RLock()
        
    def makePickle( self, record ):
        """We don't want to pickle, we want to jsonify"""
        return self.formatter.format( record )
    MAX_BUFFER = 100
    def send(self, s):
        """Send our content to the other side
        
        This implementation stores any unsent messages locally until we can 
        get to the server...
        """
        self.buffer.append( s )
        if len(self.buffer) > self.MAX_BUFFER:
            self.buffer = self.buffer[-self.MAX_BUFFER:]
        if self.sock is None:
            self.createSocket()
        if self.sock:
            while self.buffer:
                try:
                    s = self.buffer.pop(0)
                except IndexError:
                    break
                else:
                    try:
                        sentsofar = 0
                        left = len(s)
                        while left > 0:
                            sent = self.sock.send(s[sentsofar:])
                            sentsofar = sentsofar + sent
                            left = left - sent
                    except socket.error:
                        # s was not sent in full, so re-schedule...
                        self.buffer.insert(0,s)
                        self.sock.close()
                        self.sock = None  # so we can call createSocket next time

class JSONFormatter( logging.Formatter ):
    suppress = set([
        'relativeCreated','process','threadName','thread','msecs','processName',
        'exc_info','exc_text','args','message','request',
    ])
    MSG_FORMATTER = logging.Formatter()
    DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
    def format( self, record ):
        """Format the record as a subset of fields in JSON notation
        
        self.MSG_FORMATTER is applied to the record to produce "msg", so that 
        records do not need args to be transmitted
        
        Each field in self.suppress is removed from the records (mostly to 
        reduce record size).
        
        If traceback.format_exc() returns something other than ``'None\n'`` then 
        the traceback is added to the record as 'traceback' key
        
        The JSON is generated with sort_keys, so the format should be fairly 
        consistent in the final file.
        """
        raw = record.__dict__.copy()
        # Do a basic message format...
        try:
            raw['msg'] = self.MSG_FORMATTER.format( record )
        except Exception:
            raw['msg'] = 'Unable to format message'
            pass
        
        if raw['levelno'] >= logging.WARNING:
            trace = traceback.format_exc()
            if trace and trace != 'None\n':
                raw['traceback'] = trace
        raw['created_formatted'] = time.strftime(
            self.DATE_FORMAT,
            time.localtime(raw['created'])
        )
        for key in self.suppress:
            if key in raw:
                del raw[key]
        return json.dumps( raw, sort_keys=True, skipkeys=True, separators=(',', ':') ) + '\n'

def sample_log():
    handler = JSONSocketHandler( )
    handler.setLevel(logging.WARNING)
    log = logging.getLogger( 'blue' )
    log.addHandler( handler )
    log.propagate = False
    log.info( 'Should not show up' )
    log.warn( 'Regular log should show up: %s', 'this' )
    try:
        raise AttributeError( 'blue' )
    except Exception:
        log.error( 'Hello world: %s', 'this' )
        print('done')
    handler.close()
    
if __name__ == '__main__':
    sample_log()



#! /usr/bin/env python
"""Install a given firmware package onto the file system (commands fussy-install and fussy-clean)

Major TODO items:

* TODO: error handling for all of the big issues (disk-space, memory, script failures)
"""
import subprocess, os, shutil, glob, logging, traceback
from optparse import OptionParser
from fussy import unpack, errors, nbio
log = logging.getLogger( __name__ )
from ._shims import bytes, unicode

CURRENT_LINK = 'current'
FAILSAFE_NAME = 'failsafe'
PROTECTED = [CURRENT_LINK,FAILSAFE_NAME]
DEFAULT_FIRMWARE_DIRECTORY = '/opt/firmware'
LOG_FORMAT = '%(levelname)- 7s %(name)-30s %(asctime)s  %(message)s'

def clean( target=DEFAULT_FIRMWARE_DIRECTORY, protected = None ):
    """Naive cleaning implementation
    
    Removes all names in target which are not in protected,
    paths in protected must be *full* path names, as returned 
    by glob.  The target of the current link is protected.
    """
    if protected is None:
        protected = [
            os.path.join( target, p ) 
            for p in PROTECTED
        ]
    # current target is always protected...
    target = final_path(target)
    current = os.path.join( target, CURRENT_LINK)
    assert os.path.exists( current ), "Current link appears to be missing, corrupt installation  (e.g. run install)?"
    current_target = final_path( current )
    assert os.path.exists( current_target ), "Current link appears to be broken, fix before cleaning (e.g. run install)"
    protected.append( current_target )
    for path in glob.glob( os.path.join( target, '*' )):
        if path not in protected:
            log.warn( 'Removing unused firmware: %s', path )
            shutil.rmtree( path, True )
    return current_target

def final_path( link ):
    """Get the final path of the given link
    
    raises IOError if link target does not exist, or the target is not a directory 
    
    returns normalized real target path of the link
    """
    real = os.path.normpath( os.path.realpath( link ) )
    if not os.path.exists( real ):
        raise IOError( "Target of link %(link)r (%(real)s) does not exist"%locals())
    if not os.path.isdir( real ):
        raise IOError( "Target of link %(link)r is not a directory"%locals())
    return real 

def swap_link( final_target, current ):
    """Swap current link to point to final_target
    
    Steps taken:
    
        * if there is an existing tmp link, remove it 
        * create a tmp link to the final target 
        * rename tmp link to `current`
    
    returns None
    """
    tmp = current + '~'
    try:
        os.remove( tmp )
        # TODO: address race condition, you should have the while upgrade 
        # cron-locked, but that's not obvious here...
    except (OSError,IOError):
        pass
    os.symlink( os.path.basename(final_target), tmp )
    os.rename( tmp, current )

def install_bytes( filename, keyring='/etc/fussy/keys', target='/opt/firmware' ):
    """Install the packaged bytes into a final target directory
    
    Steps taken:
    
        * unpack firmware using :func:`fussy.unpack.unpack`
        * rsync new_firmware into /opt/firmware (`target`)
            * if `CURRENT_LINK` (current) is present in `target`, 
              will hard-link shared files between the new firmware and `current` 
              to reduce disk use (using :command:`rsync` parameter --link-dest)
        * removes the temporary directory where unpacking was performed
    
    returns full path to sub-directory of target where new firmware was installed
    
    raises Errors on most failures, including disk-full, failed commands, missing 
    executables, etc
    """
    temp_dir = unpack.unpack( filename, keyring )
    assert os.path.exists( temp_dir )
    base_name = os.path.basename( temp_dir )
    try:
        final_target = os.path.join( os.path.normpath( target ), base_name )
        i=0
        while os.path.exists( final_target ):
            i+= 1
            final_target = os.path.join( os.path.normpath( target ), base_name + '-%i'%(i,) )
        current = os.path.join( os.path.normpath( target ), 'current' )
        command = [
            'rsync', '-aq',
        ]
        if os.path.exists( current ):
            log.info( 'Reducing firmware size with hard-link compression' )
            command.append( '--link-dest=%s'%( current,),)
        # TODO: figure out some way to configure rsync to not create a second 
        # level directory when told `rsync -a a b`
        command.extend([
            os.path.join( temp_dir, x )
            for x in os.listdir( temp_dir )
        ])
        command.extend([
            final_target,
        ])
        log.info( 'Fixating firmware' )
        subprocess.check_call( command )
        return final_target
    finally:
        shutil.rmtree( temp_dir, True )

def enable( final_target, current, recovery=False ):
    """Attempt to enable final_target as the current release
    
    Steps taken:
    
        * runs `final_target/.pre-install final_target recovery` 
          (iff .pre-install is present)
        * (atomically) swaps the link `current` 
          for a link that points to `final_target`
        * runs `final_target/.post-install final_target recovery` 
          (iff .post-install is present)
        * if a failure occurs before swap-link completes,
          deletes final_target 
        * if a failure occurs after swap-link, raises an error
    
    returns None 
    raises Exceptions on lots of failure cases
    """
    pre_install = os.path.join( final_target, '.pre-install' )
    post_install = os.path.join( final_target, '.post-install' )
    
    try:
        if os.path.exists( pre_install ):
            log.info( 'Running pre-install script' )
            def report_progress( line ):
                log.info( 'pre-install: %s', line )
            pipe = nbio.Process( [
                pre_install,
                final_target,
                unicode(recovery).encode('utf-8'),
            ], by_line=True, stderr=-1) | report_progress
            pipe()
        log.info( 'Setting firmware current' )
        swap_link( final_target, current )
    except Exception:
        # we failed in either pre-setup or swapping
        if not recovery:
            log.warn( 'Failed during pre-install or swap, aborting' )
            shutil.rmtree( final_target, True )
            raise
        else:
            log.exception( 'Failure attempting to re-initialize previous installation, likely reboot required' )
            # Note, we will attempt to run post-install, as it is often able to perform 
            # cleanups that the pre-install cannot...
    
    if os.path.exists( post_install ):
        log.info( 'Running post-install script' )
        def report_progress( line ):
            log.info( 'post-install: %s', line )
        pipe = nbio.Process([
            post_install,
            final_target,
            unicode(recovery).encode('utf-8'),
        ], by_line=True, stderr=-1) | report_progress
        pipe()

def ensure_current_link( current, failsafe ):
    """Ensure that current is a link (not a directory)"""
    if (
        os.path.isdir( current ) and 
        not os.path.islink( current)
    ):
        if not os.path.exists(failsafe):
            # this is the initial install case, where we want failsafe to be 
            # created from the initial installation image...
            os.rename( current, failsafe )
            os.symlink( os.path.basename(failsafe), current )
        else:
            raise RuntimeError(
                """%(current)r is a directory, it should be a link.  Failsafe already exists."""%locals()
            )

def install( filename, keyring='/etc/fussy/keys', target='/opt/firmware', unpacked=False ):
    """Install given firmware <filename> into given target directory
    
    Steps taken:

        * unpack firmware (using :func:`fussy.install.install_bytes`)
        * enable firmware (using :func:`fussy.install.enable`)
        * if :func:enable fails, enable `previous` 
          (or `failsafe` if there was no previous)
    
    returns (error_code (0 is success), path name of the installed package)
    """
    current = os.path.join( os.path.normpath( target ), CURRENT_LINK )
    failsafe = os.path.join( os.path.normpath( target ), FAILSAFE_NAME )
    
    ensure_current_link( current, failsafe )
    
    previous = None
    try:
        previous = final_path( current )
    except IOError as err:
        log.warn( "Target of current does not appear to exist: %s", err )
    if not previous and os.path.exists( failsafe ):
        previous = failsafe
    log.info( 'Previous installation: %s', previous )
    log.info( 'Unpacking firmware to disk' )
    if not unpacked:
        final_target = install_bytes( filename, keyring, target )
    else:
        final_target = os.path.normpath(os.path.abspath(filename))
        if not final_target.startswith(target):
            raise RuntimeError("Firmware is not unpacked into the target directory")
        if os.path.basename(final_target) == 'current':
            raise RuntimeError("Firmware is named 'current', cannot activate")
    log.info( 'New installation: %s', final_target )
    assert os.path.exists( final_target )
    try:
        enable( final_target, current )
    except Exception as err:
        log.error( "Failure installing %s: %s", final_target, err )
        log.error( "Traceback: %s", traceback.format_exc())
        if previous:
            log.warn( "Attempting to restore previous: %s", previous )
            enable( previous, current, recovery=True )
            raise errors.RevertedFailure( previous )
        log.error( "Unable to recover, contact support!" )
        raise errors.UnrecoverableError( unicode(err) )
        # TODO: need lots more logic in the back-off code...
    else:
        log.info( "Successfully installed %s", final_target )
        return final_target

def get_options():
    """Creates the OptionParser used in :func:`main` """
    parser = OptionParser()
    parser.add_option(
        '-f','--file',
        dest = 'file',
        default = None,
        action="store",
        type="string",
        help="The firmware archive to unpack, must be a .tar.gz.gpg or a .tar.gz.asc",
    )
    parser.add_option(
        '-k','--keyring',
        dest = 'keyring',
        default = unpack.DEFAULT_KEYRING,
        action="store",
        type="string",
        help="GPG keyring to use for verification/decryption (default /etc/fussy/keys)",
    )
    parser.add_option(
        '-t','--target',
        dest = 'target',
        default = DEFAULT_FIRMWARE_DIRECTORY,
        action="store",
        type="string",
        help="Directory into which to rsync the firmware (default /opt/firmware)",
    )
    parser.add_option(
        '-u','--unpacked',
        dest = 'unpacked',
        default = False,
        action="store_true",
        help="Target is pre-unpacked image directory (fussy-install-bytes)",
    )
    parser.add_option(
        '-l', '--logfile',
        dest='logfile',
        default='/tmp/fussy-install.log',
        help = "File into which to write the fussy installation log (default /tmp/fussy-install.log)",
    )
    return parser

def configure_log( logfile ):
    if logfile:
        logging.basicConfig(filename=logfile,level=logging.INFO, format=LOG_FORMAT,datefmt='%Y-%m-%d %H:%M:%S')
    else:
        logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)
    return logfile

def main():
    """Main entry-point for the fussy-install script
    
    Steps taken:
    
        * parses arguments 
        * launches :func:`install`
    """
    parser = get_options()
    options,args = parser.parse_args()
    if not options.file:
        if args:
            options.file = args[0]
        else:
            parser.error( "Need a file to install" )
    configure_log( options.logfile )
    try:
        install( options.file, options.keyring, options.target )
        return 0
    except Exception as err:
        log.error(
            "Failure during installation: %s", err,
        )
        log.error(
            "Traceback: %s", traceback.format_exc(),
        )
        raise

def clean_main():
    """Main entry-point for fussy-clean script 
    
    Steps taken:
    
        * parses arguments 
        * launches :func:`clean`
    """
    parser = OptionParser()
    parser.add_option(
        '-t','--target',
        dest = 'target',
        default = DEFAULT_FIRMWARE_DIRECTORY,
        action="store",
        type="string",
        help="Directory into which to rsync the firmware (default /opt/firmware)",
    )
    parser.add_option(
        '-l', '--logfile',
        dest='logfile',
        default='/tmp/fussy-install.log',
        help = "File into which to write the fussy installation log (default /tmp/fussy-install.log)",
    )
    options,args = parser.parse_args()
    configure_log( options.logfile )
    clean( options.target )
    return 0

def install_bytes_main_options():
    """Creates the OptionParser used in :func:`unpack_main` """
    parser = OptionParser("Unpacks a firmware onto disk without activating it")
    parser.add_option(
        '-f','--file',
        dest = 'file',
        default = None,
        action="store",
        type="string",
        help="The firmware archive to unpack, must be a .tar.gz.gpg or a .tar.gz.asc",
    )
    parser.add_option(
        '-k','--keyring',
        dest = 'keyring',
        default = unpack.DEFAULT_KEYRING,
        action="store",
        type="string",
        help="GPG keyring to use for verification/decryption (default /etc/fussy/keys)",
    )
    parser.add_option(
        '-t','--target',
        dest = 'target',
        default = DEFAULT_FIRMWARE_DIRECTORY,
        action="store",
        type="string",
        help="Directory into which to rsync the firmware (default /opt/firmware)",
    )
    parser.add_option(
        '-l', '--logfile',
        dest='logfile',
        default='/tmp/fussy-install.log',
        help = "File into which to write the fussy installation log (default /tmp/fussy-install.log)",
    )
    return parser
def install_bytes_main():
    parser = install_bytes_main_options()
    options, args = parser.parse_args()
    if not options.file:
        if args:
            options.file = args[0]
        else:
            parser.error( "Need a file to install" )
    configure_log( options.logfile )
    try:
        final_target = install_bytes_main( options.file, options.keyring, options.target )
        print(final_target)
        return 0
    except Exception as err:
        log.error(
            "Failure during unpacking: %s", err,
        )
        log.error(
            "Traceback: %s", traceback.format_exc(),
        )
        raise
    



try:
    unicode
except NameError:
    unicode = str 
else:
    unicode = unicode
try:
    bytes 
except NameError:
    bytes = str
else:
    bytes = bytes
try:
    next
except NameError:
    def next( it ):
        return it.next()
else:
    next = next
try:
    file 
except NameError:
    import io
    file_base = io.IOBase
else:
    file_base = file
try:
    long
except NameError:
    long = int 
else:
    long = long 



#! /usr/bin/env python
from setuptools import setup, find_packages
import os

def get_version():
    filename = os.path.join('fussy','version.py') 
    if not os.path.exists(filename):
        return '1.0.0'
    for line in open(filename):
        if line.startswith( '__version__' ):
            return line.split('=')[1].strip().strip('"').strip("'")
    raise RuntimeError( "Unable to determine version" )

if __name__ == "__main__":
    setup(
        name='fussy',
        version=get_version(),
        description='Field-Upgradable Software Support',
        long_description='''Fussy Software Packager

Fussy provides a collection of tools to create field-upgraded
software installations. The installation uses a "pivot" root
on your regular file system into which the current release of 
your software is installed, and which can exist alongside a
number of other versions of the software.

It assumes a Linux-like operating system and requires certain
linux services and utilities for many of its functions.

Notable functionality:

* GPG signed/verified firmware packing and unpacking
* inotify support
* non-blocking process IO (nbio)
* transactional write operation
* process control operations (e.g. kill-kill-kill)
* simple structured log server
* cron locks
''',
        classifiers=[
            "Programming Language :: Python",
            "Operating System :: POSIX :: Linux",
            "License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)",
        ],
        license="LGPL",
        author='Mike C. Fletcher',
        author_email='mcfletch@vrplumber.com',
        url='http://www.vrplumber.com',
        keywords='firmware,field,upgrade,embedding',
        packages=find_packages(),
        include_package_data=True,
        # system-level requirements 
        # python2.6+
        # gpg
        # rsync
        # tar
        install_requires=[
            'setuptools',
        ],
        entry_points = dict(
            console_scripts = [
                'fussy-install=fussy.install:main',
                'fussy-install-bytes=fussy.install:install_bytes_main', 
                'fussy-clean=fussy.install:clean_main',
                'fussy-unpack=fussy.unpack:main',
                'fussy-pack=fussy.pack:main',
                'fussy-sign=fussy.pack:sign_main',
                'fussy-south-rollback=fussy.southrollback:main',
                'fussy-link-tree=fussy.linktree:main',
                'fussy-log-server=fussy.jsonlogserver:main',
                'fussy-format-log=fussy.jsonlogserver:format_log_main',
                'json-format=fussy.jsonformat:main', 
                'fussy-file-watch=fussy.filewatch:main',
                'fussy-jsx-watcher=fussy.jsxwatcher:main',
                'fussy-pack-unattended=fussy.unattendedset:main', 
                'fussy-deb-metadata=fussy.unattendedset:metadata_main',
                'fussy-sse-server=fussy.sseserver:main', 
                'fussy-sse-client-demo=fussy.sseclient:main',
            ],
        ),
    )



#! /usr/bin/env python
"""Bundle a package as a signed firmware for installation/redistribution
"""
import tempfile, subprocess, os
from optparse import OptionParser

DEFAULT_EXCLUDES = [
    '.svn',
    '.bzr',
    '.git',
]

def pack( root_dir, excludes=None ):
    """Bundle directory into a firmware file...
    
    * root_dir -- directory to be packed into the firmware, the 
        :func:`os.path.basename` of this directory will be the name used 
        in the created bundle filename and the directory installed into 
        the target on client machines
    * excludes -- patterns to exclude from the firmware image (passed to
        :command:`tar` during packing)
    
    returns absolute filename for generated .tar.gz firmware 
    (created in a temporary directory)
    """
    temp_dir = tempfile.mkdtemp(prefix='fussy-',suffix='-pack')
    firmware_name = os.path.basename( root_dir )
    if firmware_name == 'current':
        raise RuntimeError( 'Cannot create a firmware named "current"')
    tar_name = firmware_name+ '.tar.gz'
    tar_file = os.path.join( temp_dir, tar_name )
    command = [
        'tar','-zcf',
        tar_file,
    ] + [
        '--exclude=%s'%(x,) 
        for x in ((excludes or []) + DEFAULT_EXCLUDES)
    ] + [os.path.basename(root_dir)]
    subprocess.check_call( command, cwd=os.path.dirname( os.path.abspath(root_dir) ) )
    return tar_file

def sign( tar_file, key=None,encrypt_for=None ):
    """Sign an existing tar.gz-file with the default key
    
    * tar_file -- tar.gz file to sign (and/or encrypt)
    * encrypt_for -- if non-null, string identifier for which to encrypt
    """
    gpg_file = tar_file + '.gpg'
    command = [ 'gpg' ]
    if encrypt_for:
        command += ['-se', '-r', encrypt_for]
    else:
        command += ['--sign']
    if key:
        command += ['-u',key]
    command.append( tar_file )
    subprocess.check_call( command )
    assert os.path.exists( gpg_file )
    os.remove( tar_file )
    return gpg_file
    

def get_options():
    """Produces the OptionParser for :func:main"""
    parser = OptionParser()
    parser.add_option( 
        '-x','--exclude', 
        dest='exclude', 
        action="append",
        type="string",
        help="Paths/patterns to exclude from the archive",
    )
    parser.add_option(
        '-r','--root',
        dest = 'root',
        default='firmware',
        type="string",
        help="The directory to be packed for distribution, should be a versioned/unique name, for instance including a human-readable timestamp",
    )
    parser.add_option(
        '-u','--unsigned',
        dest='signed',
        action="store_false",
        default=True,
        help="If specified, do not do signing, just produced an unsigned tar.gz file (intended to allow for signing on a separate workstation)",
    )
    add_encryption_options( parser )
    return parser
def add_encryption_options( parser ):
    parser.add_option(
        '-k','--key',
        dest='key',
        default = None,
        help = 'If specified, use this key id to perform the signing (passed to :cmd:gpg -u option), otherwise uses default identity',
    )
    parser.add_option(
        '-e','--encrypt-for',
        dest = 'encrypt',
        action="store",
        type="string",
        help="The name of the key for which to encrypt (otherwise just sign)",
    )
    return parser
    
def main():
    """Main function for the packing script"""
    options,args = get_options().parse_args()
    tar_file = pack( options.root, options.exclude )
    if options.signed:
        gpg_file = sign( tar_file, key=options.key, encrypt_for=options.encrypt )
        print(gpg_file) 
    else:
        print(tar_file)
    return 0

def get_sign_options():
    parser = OptionParser()
    parser.add_option(
        '-f','--filename',
        dest='filename',
        default = None,
        help = 'The full path to the tar.gz file we will sign',
    )
    add_encryption_options( parser )
    return parser

def sign_main():
    """Main function that *just* signs a tar-file"""
    options,args = get_sign_options().parse_args()
    if not options.filename and args:
        options.filename = args[0]
    gpg_file = sign( options.filename, key=options.key, encrypt_for=options.encrypt )
    print(gpg_file)
    return 0
    



import logging, re, os, fnmatch
from . import inotify 
log = logging.getLogger(__name__)

class FileWatcher(object):
    """Watches a directory for files matching fnmatch
    
    Assumes that all file-writes will be done with twrite,
    and by default assumes that ~ suffixed files should not 
    be processed (again, the twrite default temp-file suffix).
    
    .. doctest::
    
        >>> import os,shutil,tempfile
        >>> from fussy import twrite
        >>> temp = tempfile.mkdtemp(prefix='filewatcher-',suffix='-test')
        >>> with open(os.path.join(temp,'test.moo'),'wb') as fh:
        ...    written = fh.write(b'a')
        >>> f = FileWatcher( temp, recursive=True )
        >>> for event in f.start():
        ...    print(event) # doctest: +ELLIPSIS
        InotifyEvent(...,...,IN_CREATE,0,'.../filewatcher-...-test/test.moo')
        >>> with open(os.path.join(temp,'test2.txt'),'wb') as fh:
        ...    written = fh.write(b'a')
        >>> for event in f.iterate_events(None): # do not block
        ...    print(event) # doctest: +ELLIPSIS
        InotifyEvent(...,...,IN_CREATE,0,'.../filewatcher-...-test/test2.txt')
        >>> twrite.twrite( os.path.join(temp,'ieee','zoom'),'content')
        >>> for event in f.iterate_events(None): # do not block
        ...    print(event) # doctest: +ELLIPSIS
        InotifyEvent(...,...,IN_CREATE|IN_ISDIR,0,'.../filewatcher-...-test/ieee')
        InotifyEvent(...,...,IN_CREATE,0,'.../filewatcher-...-test/ieee/zoom')
        >>> shutil.rmtree( os.path.join(temp,'ieee'))
        >>> for event in f.iterate_events(None): # do not block
        ...    print(event) # doctest: +ELLIPSIS
        InotifyEvent(...,...,IN_DELETE,0,'.../filewatcher-...-test/ieee/zoom')
        InotifyEvent(...,...,IN_DELETE|IN_ISDIR,0,'.../filewatcher-...-test/ieee')
        >>> f.unwatch_directory( '/moo' ) # won't do anything, as it's not watched...
        >>> f.close()
        >>> f2 = FileWatcher(temp,pattern='*.moo',recursive=False)
        >>> for event in f2.start():
        ...    print(event) # doctest: +ELLIPSIS
        InotifyEvent(...,...,IN_CREATE,0,'.../filewatcher-...-test/test.moo')
        >>> shutil.rmtree(temp)
    
    """
    # maximum time before we iterate to avoid infinite hangs
    POLL_PERIOD = 20
    DEFAULT_EVENTS = (
        inotify.IN_MOVE|inotify.IN_MODIFY|
        inotify.IN_CREATE|inotify.IN_DELETE|
        inotify.IN_MOVED_TO|inotify.IN_MOVED_FROM
    )
    active = False
    def __init__(
        self,
        directory,
        pattern=re.compile(r'.+[^~]$'),
        events = None,
        recursive = False,
    ):
        """Watch directory for files matching pattern
        
        directory -- path to a directory
        pattern -- fnmatch string *or* regex object for matching
        """
        self.pattern = pattern
        self.directories = [directory]
        self.watches = {}
        self.back_watches = {}
        if events is None:
            events = self.DEFAULT_EVENTS
        self.events = events
        self.recursive = recursive
    
    def match(self, filename ):
        """Does this filename match our patterns?"""
        if hasattr(self.pattern,'match'):
            return self.pattern.match(filename)
        else:
            return fnmatch.fnmatchcase(os.path.basename(filename),self.pattern)
    
    _fd = None
    @property
    def fd(self):
        if self._fd is None:
            self._fd = inotify.Inotify()
        return self._fd
    def fileno(self):
        return self.fd
    
    def watch_directory(self,directory):
        if directory not in self.directories:
            self.directories.append(directory)
        if self.active and directory not in self.watches:
            log.info('Start watching: %s', directory)
            wd = self.watches[directory] = self.fd.add_watch(
                directory, 
                self.events,
            )
            log.debug('Watch: %s -> %s,%s',directory,int(self.fd),int(wd))
            return wd
    def unwatch_directory(self,directory,deleted=False):
        """De-register watches on given directory"""
        try:
            log.info('Stop watching: %s', directory)
            wd = self.watches.pop(directory)
        except KeyError:
            pass 
        else:
            try:
                self.fd.rm_watch(wd)
            except Exception as err:
                if hasattr(err,'errno') and err.errno == 22:
                    pass 
                elif not deleted:
                    log.warn('Failure during removal %s: %s', err,directory)
        while directory in self.directories:
            self.directories.remove(directory)
        prefix = directory + os.path.sep
        if prefix.endswith( os.path.sep *2 ):
            prefix = prefix[:-1]
        log.debug("Checking for prefix: %s",prefix)
        for subdir in [
            s for s in self.directories 
            if s.startswith( prefix )
        ]:
            self.unwatch_directory(subdir)
        
    def __iter__(self):
        log.info('Starting file watcher')
        try:
            for event in self.start():
                yield event
            while self.active:
                for event in self.iterate_events(self.POLL_PERIOD):
                    yield event
        finally:
            self.close()
    def start(self):
        """Start up our iterations
        
        yields create events for each matching file...
        """
        self.active = True
        for event in self.initial_watches():
            yield event
    def close(self):
        """Close our watches and shut down"""
        fd = self._fd 
        if fd:
            fd.close()
        self.watches.clear()
        self.active = False
        
    def initial_watches(self):
        """Walk our directories adding them to our watcher"""
        for directory in self.directories:
            for match in self.on_directory_added(directory):
                yield match
    def iterate_events(self,period=None):
        """Do a single iteration of polling for events for a given period (blocking)
        
        This interface is intended to be used from a thread, so that you 
        get a simple iterable producing events to be processed. The processing
        for each event should be relatively lightweight, as it needs to 
        return fast enough to avoid having the event queue back up.
        """
        events = self.fd.get_events(block=True,period=period or self.POLL_PERIOD)
        events = [
            e for e in events 
            if not (
                (e.mask & inotify.IN_IGNORED) or (e.mask & inotify.IN_DELETE_SELF)
            )
        ]
        for event in events:
            fullname = event.full_name
            if event.isdir:
                yield event
                if self.recursive:
                    if event.mask & inotify.IN_DELETE or event.mask & inotify.IN_MOVED_FROM:
                        self.on_directory_deleted( fullname )
                    elif event.mask & inotify.IN_CREATE or event.mask & inotify.IN_MOVED_TO:
                        for match in self.on_directory_added( fullname ):
                            yield match
            elif event.mask & inotify.IN_DELETE_SELF:
                log.debug('Watched directory removed: %s',fullname )
            elif self.match( fullname ):
                yield event
        
    
    def on_directory_deleted(self,directory):
        """A directory has been deleted"""
        self.unwatch_directory( directory, deleted=True )
    def on_directory_added(self,directory):
        """A directory has been created"""
        if self.recursive:
            for base,subdirs,files in os.walk(directory):
                wd = self.watch_directory( os.path.join(directory,base) )
                if wd:
                    for file in files:
                        fullname = os.path.join(base,file)
                        if self.match(fullname):
                            yield inotify.InotifyEvent(
                                fd = self.fd,
                                wd = wd,
                                mask=inotify.IN_CREATE,
                                cookie=0,
                                name=file,
                            )
        else:
            wd = self.watch_directory( directory )
            if wd:
                for file in os.listdir(directory):
                    fullname = os.path.join(directory,file)
                    if self.match(fullname):
                        yield inotify.InotifyEvent(
                            fd = self.fd,
                            wd = wd,
                            mask=inotify.IN_CREATE,
                            cookie=0,
                            name=file,
                        )

def options():
    """Create the options for demonstration watcher
    
    .. doctest::
    
        >>> parser = options()
        >>> args = parser.parse_args(['--recursive','/tmp/test'])
        >>> args.recursive
        True
        >>> args.target
        '/tmp/test'
    
    """
    import argparse
    parser = argparse.ArgumentParser(description='Watch directory for files matching given regex')
    parser.add_argument(
        '--recursive', action='store_true',
        default=False,
        help='Recursively watch all directories under the directory'
    )
    parser.add_argument(
        'target', metavar='DIR',
        help="Target directory to watch (must exist)",
    )
    return parser
    
def main():
    logging.basicConfig(level=logging.INFO)
    args = options().parse_args()
    r = FileWatcher( args.target, recursive=args.recursive, )
    for event in r:
        log.info('%r', event)
